# Familiar - Self-Hosted AI Companion Platform
# Copyright (c) 2026 George Scott Foley
# ORCID: 0009-0006-4957-0540
# Licensed under the MIT License - see LICENSE file for details

"""
Tests for the self_management skill.

Covers: show_self_status, list_config_checkpoints, propose_config_change,
rollback_config, apply_update, restart_agent, and TOOLS registry.
"""

import json
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import yaml

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


@pytest.fixture
def tmp_config(tmp_path):
    """Redirect config and checkpoint paths to tmp_path."""
    config_file = tmp_path / "config.yaml"
    checkpoints_dir = tmp_path / "config_checkpoints"
    config_file.write_text(
        yaml.dump(
            {
                "agent": {
                    "name": "TestAgent",
                    "persona": "a test assistant",
                    "security_mode": "balanced",
                    "allow_shell_commands": False,
                    "scheduler_enabled": True,
                },
                "llm": {"provider": "ollama", "model": "llama3.2"},
            }
        )
    )
    with (
        patch("familiar.skills.self_management.skill.CONFIG_FILE", config_file),
        patch("familiar.skills.self_management.skill.CHECKPOINTS_DIR", checkpoints_dir),
        patch("familiar.skills.self_management.skill.DATA_DIR", tmp_path),
    ):
        yield config_file, checkpoints_dir


# ---------------------------------------------------------------------------
# show_self_status
# ---------------------------------------------------------------------------


class TestShowSelfStatus:
    def test_returns_version(self, tmp_config):
        from familiar.skills.self_management.skill import show_self_status

        result = show_self_status({})
        assert "version" in result

    def test_returns_platform_info(self, tmp_config):
        from familiar.skills.self_management.skill import show_self_status

        result = show_self_status({})
        assert "python_version" in result
        assert "hostname" in result
        assert "platform" in result

    def test_returns_config_summary(self, tmp_config):
        from familiar.skills.self_management.skill import show_self_status

        result = show_self_status({})
        assert "config_summary" in result
        summary = result["config_summary"]
        assert summary["agent_name"] == "TestAgent"
        assert summary["llm_model"] == "llama3.2"

    def test_returns_config_file_path(self, tmp_config):
        from familiar.skills.self_management.skill import show_self_status

        result = show_self_status({})
        assert "config_file" in result
        assert "config_exists" in result

    def test_loaded_skill_count_non_negative(self, tmp_config):
        from familiar.skills.self_management.skill import show_self_status

        result = show_self_status({})
        assert result.get("loaded_skill_count", 0) >= 0

    def test_checkpoint_count_zero_initially(self, tmp_config):
        from familiar.skills.self_management.skill import show_self_status

        result = show_self_status({})
        assert result["config_checkpoints"] == 0


# ---------------------------------------------------------------------------
# list_config_checkpoints
# ---------------------------------------------------------------------------


class TestListConfigCheckpoints:
    def test_empty_initially(self, tmp_config):
        from familiar.skills.self_management.skill import list_config_checkpoints

        result = list_config_checkpoints({})
        assert result["checkpoints"] == []
        assert result["count"] == 0

    def test_lists_saved_checkpoints(self, tmp_config):
        config_file, checkpoints_dir = tmp_config
        checkpoints_dir.mkdir(parents=True, exist_ok=True)

        # Manually write a checkpoint
        ck = {
            "id": "20260312_120000",
            "reason": "test checkpoint",
            "timestamp": "2026-03-12T12:00:00+00:00",
            "config": {"agent": {"name": "Old"}},
        }
        (checkpoints_dir / "20260312_120000.json").write_text(json.dumps(ck))

        from familiar.skills.self_management.skill import list_config_checkpoints

        result = list_config_checkpoints({})
        assert result["count"] == 1
        assert result["checkpoints"][0]["id"] == "20260312_120000"
        assert result["checkpoints"][0]["reason"] == "test checkpoint"

    def test_newest_first(self, tmp_config):
        config_file, checkpoints_dir = tmp_config
        checkpoints_dir.mkdir(parents=True, exist_ok=True)

        for ts in ("20260312_100000", "20260312_110000", "20260312_120000"):
            ck = {"id": ts, "reason": "", "timestamp": ts, "config": {}}
            (checkpoints_dir / f"{ts}.json").write_text(json.dumps(ck))

        from familiar.skills.self_management.skill import list_config_checkpoints

        result = list_config_checkpoints({})
        ids = [c["id"] for c in result["checkpoints"]]
        assert ids == sorted(ids, reverse=True)


# ---------------------------------------------------------------------------
# propose_config_change
# ---------------------------------------------------------------------------


class TestProposeConfigChange:
    def test_requires_key(self, tmp_config):
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change({"value": "X", "reason": "test"})
        assert "error" in result

    def test_requires_value(self, tmp_config):
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change({"key": "agent.name", "reason": "test"})
        assert "error" in result

    def test_requires_reason(self, tmp_config):
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change({"key": "agent.name", "value": "X"})
        assert "error" in result

    def test_blocks_non_editable_key(self, tmp_config):
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change(
            {"key": "agent.internal_secret", "value": "x", "reason": "testing"}
        )
        assert "error" in result
        assert "not in the list" in result["error"]

    def test_returns_confirmation_sentinel(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change(
            {"key": "agent.name", "value": "NewName", "reason": "renaming agent"}
        )
        assert isinstance(result, ConfirmationRequired)

    def test_confirmation_preview_shows_diff(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change(
            {"key": "agent.name", "value": "NewName", "reason": "renaming agent"}
        )
        assert isinstance(result, ConfirmationRequired)
        assert "TestAgent" in result.preview  # current value
        assert "NewName" in result.preview  # proposed value
        assert "renaming agent" in result.preview  # reason

    def test_applies_change_when_confirmed(self, tmp_config):
        config_file, checkpoints_dir = tmp_config
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change(
            {
                "key": "agent.name",
                "value": "Confirmed",
                "reason": "test apply",
                "_confirmed": True,
            }
        )
        assert result.get("applied") is True
        assert result["new_value"] == "Confirmed"
        assert result["old_value"] == "TestAgent"

        # Verify config file was actually written
        with open(config_file) as f:
            written = yaml.safe_load(f)
        assert written["agent"]["name"] == "Confirmed"

    def test_saves_checkpoint_before_applying(self, tmp_config):
        config_file, checkpoints_dir = tmp_config
        checkpoints_dir.mkdir(parents=True, exist_ok=True)

        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change(
            {
                "key": "agent.name",
                "value": "Updated",
                "reason": "checkpoint test",
                "_confirmed": True,
            }
        )
        assert "checkpoint_id" in result
        # Checkpoint file should exist
        ck_path = checkpoints_dir / f"{result['checkpoint_id']}.json"
        assert ck_path.exists()

        # Checkpoint should contain the OLD config
        with open(ck_path) as f:
            ck = json.load(f)
        assert ck["config"]["agent"]["name"] == "TestAgent"


# ---------------------------------------------------------------------------
# rollback_config
# ---------------------------------------------------------------------------


class TestRollbackConfig:
    def test_requires_checkpoint_id(self, tmp_config):
        from familiar.skills.self_management.skill import rollback_config

        result = rollback_config({})
        assert "error" in result

    def test_not_found_returns_error(self, tmp_config):
        from familiar.skills.self_management.skill import rollback_config

        result = rollback_config({"checkpoint_id": "nonexistent_123"})
        assert "error" in result

    def test_returns_confirmation_sentinel(self, tmp_config):
        config_file, checkpoints_dir = tmp_config
        checkpoints_dir.mkdir(parents=True, exist_ok=True)

        ck = {
            "id": "20260312_120000",
            "reason": "before change",
            "timestamp": "2026-03-12T12:00:00+00:00",
            "config": {"agent": {"name": "OldName"}},
        }
        (checkpoints_dir / "20260312_120000.json").write_text(json.dumps(ck))

        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import rollback_config

        result = rollback_config({"checkpoint_id": "20260312_120000"})
        assert isinstance(result, ConfirmationRequired)
        assert result.risk == "high"

    def test_restores_config_when_confirmed(self, tmp_config):
        config_file, checkpoints_dir = tmp_config
        checkpoints_dir.mkdir(parents=True, exist_ok=True)

        ck = {
            "id": "20260312_120000",
            "reason": "pre-change",
            "timestamp": "2026-03-12T12:00:00+00:00",
            "config": {"agent": {"name": "OldName", "security_mode": "paranoid"}},
        }
        (checkpoints_dir / "20260312_120000.json").write_text(json.dumps(ck))

        from familiar.skills.self_management.skill import rollback_config

        result = rollback_config({"checkpoint_id": "20260312_120000", "_confirmed": True})
        assert result.get("restored") is True

        with open(config_file) as f:
            written = yaml.safe_load(f)
        assert written["agent"]["name"] == "OldName"

    def test_saves_safety_checkpoint_before_restoring(self, tmp_config):
        config_file, checkpoints_dir = tmp_config
        checkpoints_dir.mkdir(parents=True, exist_ok=True)

        ck = {
            "id": "20260312_120000",
            "reason": "test",
            "timestamp": "2026-03-12T12:00:00+00:00",
            "config": {"agent": {"name": "OldName"}},
        }
        (checkpoints_dir / "20260312_120000.json").write_text(json.dumps(ck))

        from familiar.skills.self_management.skill import rollback_config

        result = rollback_config({"checkpoint_id": "20260312_120000", "_confirmed": True})
        assert "safety_checkpoint_id" in result
        safety_path = checkpoints_dir / f"{result['safety_checkpoint_id']}.json"
        assert safety_path.exists()


# ---------------------------------------------------------------------------
# apply_update
# ---------------------------------------------------------------------------


class TestApplyUpdate:
    def test_up_to_date_no_confirmation(self, tmp_config):
        from familiar.skills.self_management.skill import apply_update

        with patch(
            "familiar.skills.self_management.skill._pip_dry_run",
            return_value=(False, "Already up to date."),
        ):
            result = apply_update({"channel": "stable"})
        assert result.get("up_to_date") is True
        # No confirmation sentinel — already up to date
        assert "message" in result

    def test_returns_confirmation_when_update_available(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import apply_update

        with patch(
            "familiar.skills.self_management.skill._pip_dry_run",
            return_value=(True, "Would install familiar-agent-1.15.99"),
        ):
            result = apply_update({"channel": "stable"})
        assert isinstance(result, ConfirmationRequired)
        assert "familiar-agent-1.15.99" in result.preview

    def test_runs_pip_on_confirmation(self, tmp_config):
        from familiar.skills.self_management.skill import apply_update

        mock_run = MagicMock()
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = "Successfully installed familiar-agent-1.15.99"
        mock_run.return_value.stderr = ""

        with patch("familiar.skills.self_management.skill.subprocess.run", mock_run):
            result = apply_update({"channel": "stable", "_confirmed": True})

        assert result.get("updated") is True
        call_args = mock_run.call_args[0][0]
        assert "install" in call_args
        assert "--upgrade" in call_args

    def test_channel_all_uses_full_extra(self, tmp_config):
        from familiar.skills.self_management.skill import apply_update

        mock_run = MagicMock()
        mock_run.return_value.returncode = 0
        mock_run.return_value.stdout = "ok"
        mock_run.return_value.stderr = ""

        with patch("familiar.skills.self_management.skill.subprocess.run", mock_run):
            apply_update({"channel": "all", "_confirmed": True})

        call_args = mock_run.call_args[0][0]
        assert any("full" in arg for arg in call_args)

    def test_pip_failure_returns_updated_false(self, tmp_config):
        from familiar.skills.self_management.skill import apply_update

        mock_run = MagicMock()
        mock_run.return_value.returncode = 1
        mock_run.return_value.stdout = ""
        mock_run.return_value.stderr = "ERROR: could not install"

        with patch("familiar.skills.self_management.skill.subprocess.run", mock_run):
            result = apply_update({"channel": "stable", "_confirmed": True})

        assert result.get("updated") is False


# ---------------------------------------------------------------------------
# restart_agent
# ---------------------------------------------------------------------------


class TestRestartAgent:
    def test_returns_confirmation_sentinel(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import restart_agent

        result = restart_agent({"reason": "applying config changes"})
        assert isinstance(result, ConfirmationRequired)

    def test_preview_shows_reason(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import restart_agent

        result = restart_agent({"reason": "new model loaded"})
        assert isinstance(result, ConfirmationRequired)
        assert "new model loaded" in result.preview

    def test_preview_shows_restart_method(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import restart_agent

        result = restart_agent({})
        assert isinstance(result, ConfirmationRequired)
        # Method should appear in preview
        assert any(m in result.preview for m in ("systemd", "supervisord", "direct"))

    def test_execv_called_when_confirmed_direct(self, tmp_config):
        from familiar.skills.self_management.skill import restart_agent

        with patch(
            "familiar.skills.self_management.skill._detect_restart_method",
            return_value="direct (os.execv)",
        ):
            with patch("familiar.skills.self_management.skill.os.execv") as mock_execv:
                restart_agent({"reason": "test", "_confirmed": True})
        mock_execv.assert_called_once()

    def test_systemctl_called_when_confirmed_systemd(self, tmp_config):
        from familiar.skills.self_management.skill import restart_agent

        with patch(
            "familiar.skills.self_management.skill._detect_restart_method",
            return_value="systemd",
        ):
            with patch("familiar.skills.self_management.skill.subprocess.Popen") as mock_popen:
                result = restart_agent({"reason": "test", "_confirmed": True})
        assert result.get("restarting") is True
        call_args = mock_popen.call_args[0][0]
        assert "systemctl" in call_args
        assert "restart" in call_args


# ---------------------------------------------------------------------------
# TOOLS registry
# ---------------------------------------------------------------------------


class TestSelfManagementToolsRegistry:
    def test_tools_list_exists(self):
        from familiar.skills.self_management.skill import TOOLS

        assert isinstance(TOOLS, list)
        assert len(TOOLS) == 6

    def test_all_tools_have_required_fields(self):
        from familiar.skills.self_management.skill import TOOLS

        for tool in TOOLS:
            assert "name" in tool
            assert "description" in tool
            assert "input_schema" in tool
            assert callable(tool["handler"])
            assert tool.get("category") == "self_management"

    def test_all_expected_tools_present(self):
        from familiar.skills.self_management.skill import TOOLS

        names = {t["name"] for t in TOOLS}
        expected = {
            "show_self_status",
            "list_config_checkpoints",
            "propose_config_change",
            "rollback_config",
            "apply_update",
            "restart_agent",
        }
        assert names == expected

    def test_read_only_tools_no_required_params(self):
        """show_self_status and list_config_checkpoints take no required params."""
        from familiar.skills.self_management.skill import TOOLS

        read_only = {
            t["name"]: t
            for t in TOOLS
            if t["name"] in ("show_self_status", "list_config_checkpoints")
        }
        for tool in read_only.values():
            assert tool["input_schema"]["required"] == []


# ---------------------------------------------------------------------------
# Editable sections guard
# ---------------------------------------------------------------------------


class TestEditableSectionsGuard:
    def test_agent_name_editable(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change({"key": "agent.name", "value": "X", "reason": "test"})
        assert isinstance(result, ConfirmationRequired)

    def test_llm_model_editable(self, tmp_config):
        from familiar.core.confirmations import ConfirmationRequired
        from familiar.skills.self_management.skill import propose_config_change

        result = propose_config_change(
            {"key": "llm.model", "value": "qwen3", "reason": "switching model"}
        )
        assert isinstance(result, ConfirmationRequired)

    def test_internal_state_not_editable(self, tmp_config):
        from familiar.skills.self_management.skill import propose_config_change

        for blocked in ("agent._internal", "llm._api_secret", "channels.discord.token"):
            result = propose_config_change({"key": blocked, "value": "x", "reason": "test"})
            assert "error" in result, f"{blocked} should be blocked"


# ---------------------------------------------------------------------------
# Job class weights
# ---------------------------------------------------------------------------


class TestJobClassWeights:
    def test_helper_has_self_management_weights(self):
        from familiar.jobs.helper import HELPER

        weights = HELPER.skill_weights
        for tool in ("show_self_status", "propose_config_change", "apply_update", "restart_agent"):
            assert tool in weights, f"{tool} missing from Helper skill_weights"

    def test_network_admin_has_self_management_weights(self):
        from familiar.jobs.network_admin import NETWORK_ADMIN

        weights = NETWORK_ADMIN.skill_weights
        for tool in ("show_self_status", "propose_config_change", "restart_agent"):
            assert tool in weights, f"{tool} missing from Network Admin skill_weights"



# Familiar - Self-Hosted AI Companion Platform
# Copyright (c) 2026 George Scott Foley
# Licensed under the MIT License

"""Nonprofit Director routes — /api/nonprofit/*, /api/bookkeeping/*, /api/legislation/*, /api/grants-gov/*."""

from __future__ import annotations

import json
import logging
import os
from datetime import datetime
from pathlib import Path

from flask import Blueprint, Response, jsonify, request, send_file

from ..auth import require_auth
from ..helpers import get_agent, DATA_DIR, safe_error

logger = logging.getLogger(__name__)

bp = Blueprint("nonprofit", __name__)


# Nonprofit CRUD API
# ============================================================


@bp.route("/api/nonprofit/donors")
@require_auth
def api_nonprofit_donors():
    """List all donors."""
    from familiar.skills.nonprofit.skill import get_all_donors

    return jsonify({"donors": get_all_donors()})


@bp.route("/api/nonprofit/donors", methods=["POST"])
@require_auth
def api_nonprofit_donor_add():
    """Create a new donor."""
    data = request.json or {}
    if not data.get("name", "").strip():
        return jsonify({"error": "Name is required."}), 400

    from familiar.skills.nonprofit.skill import log_donor

    log_donor(data)

    # Return the newly created donor
    from familiar.skills.nonprofit.skill import get_all_donors

    donors = get_all_donors()
    # Find by name match (just created)
    created = next(
        (d for d in donors if d["name"].lower() == data["name"].strip().lower()),
        donors[-1] if donors else None,
    )
    return jsonify({"donor": created}), 201


@bp.route("/api/nonprofit/donors/<donor_id>", methods=["PUT"])
@require_auth
def api_nonprofit_donor_update(donor_id):
    """Update a donor."""
    data = request.json or {}
    data["id"] = donor_id

    from familiar.skills.nonprofit.skill import update_donor

    result = update_donor(data)
    if "error" in result:
        return jsonify(result), 404
    return jsonify({"donor": result})


@bp.route("/api/nonprofit/donors/<donor_id>", methods=["DELETE"])
@require_auth
def api_nonprofit_donor_delete(donor_id):
    """Delete a donor."""
    from familiar.skills.nonprofit.skill import delete_donor

    result = delete_donor(donor_id)
    if "error" in result:
        return jsonify(result), 404
    return jsonify(result)


@bp.route("/api/nonprofit/donors/<donor_id>/interaction", methods=["POST"])
@require_auth
def api_nonprofit_donor_interaction(donor_id):
    """Log an interaction for a donor."""
    data = request.json or {}
    data["donor_id"] = donor_id

    from familiar.skills.nonprofit.skill import add_interaction

    result = add_interaction(data)
    if "error" in result:
        return jsonify(result), 404
    return jsonify({"donor": result})


@bp.route("/api/nonprofit/grants")
@require_auth
def api_nonprofit_grants():
    """List all grants."""
    from familiar.skills.nonprofit.skill import get_all_grants

    return jsonify({"grants": get_all_grants()})


@bp.route("/api/nonprofit/grants", methods=["POST"])
@require_auth
def api_nonprofit_grant_add():
    """Create a new grant."""
    data = request.json or {}
    if not data.get("name", "").strip():
        return jsonify({"error": "Grant name is required."}), 400

    from familiar.skills.nonprofit.skill import add_grant

    add_grant(data)

    from familiar.skills.nonprofit.skill import get_all_grants

    grants = get_all_grants()
    return jsonify({"grant": grants[-1] if grants else None}), 201


@bp.route("/api/nonprofit/grants/<grant_id>", methods=["PUT"])
@require_auth
def api_nonprofit_grant_update(grant_id):
    """Update a grant."""
    data = request.json or {}
    data["id"] = grant_id

    from familiar.skills.nonprofit.skill import update_grant

    result = update_grant(data)
    if "error" in result:
        return jsonify(result), 404
    return jsonify({"grant": result})


@bp.route("/api/nonprofit/grants/<grant_id>", methods=["DELETE"])
@require_auth
def api_nonprofit_grant_delete(grant_id):
    """Delete a grant."""
    from familiar.skills.nonprofit.skill import delete_grant

    result = delete_grant(grant_id)
    if "error" in result:
        return jsonify(result), 404
    return jsonify(result)


@bp.route("/api/nonprofit/notes")
@require_auth
def api_nonprofit_notes():
    """List all notes."""
    from familiar.skills.nonprofit.skill import get_all_notes

    return jsonify({"notes": get_all_notes()})


@bp.route("/api/nonprofit/notes", methods=["POST"])
@require_auth
def api_nonprofit_note_add():
    """Create a new note."""
    data = request.json or {}
    if not data.get("content", "").strip():
        return jsonify({"error": "Note content is required."}), 400

    from familiar.skills.nonprofit.skill import add_note

    add_note(data)

    from familiar.skills.nonprofit.skill import get_all_notes

    notes = get_all_notes()
    return jsonify({"note": notes[-1] if notes else None}), 201


@bp.route("/api/nonprofit/notes/<note_id>", methods=["DELETE"])
@require_auth
def api_nonprofit_note_delete(note_id):
    """Delete a note."""
    from familiar.skills.nonprofit.skill import delete_note

    result = delete_note(note_id)
    if "error" in result:
        return jsonify(result), 404
    return jsonify(result)


# ============================================================
# Bookkeeping CRUD API
# ============================================================


@bp.route("/api/bookkeeping/transactions")
@require_auth
def api_bookkeeping_transactions():
    """List transactions with optional filters."""
    from familiar.skills.bookkeeping.skill import get_all_transactions

    filters = {
        "start_date": request.args.get("start_date"),
        "end_date": request.args.get("end_date"),
        "category": request.args.get("category"),
        "fund": request.args.get("fund"),
        "type": request.args.get("type"),
    }
    # Remove None values
    filters = {k: v for k, v in filters.items() if v}
    txns = get_all_transactions(filters if filters else None)
    limit = request.args.get("limit", 200, type=int)
    return jsonify({"transactions": txns[:limit]})


@bp.route("/api/bookkeeping/transactions", methods=["POST"])
@require_auth
def api_bookkeeping_transaction_add():
    """Create a new transaction (income or expense)."""
    data = request.json or {}
    if not data.get("amount"):
        return jsonify({"error": "Amount is required."}), 400

    tx_type = data.get("type", "income")
    if tx_type == "expense":
        from familiar.skills.bookkeeping.skill import log_expense

        log_expense(data)
    else:
        from familiar.skills.bookkeeping.skill import log_income

        log_income(data)

    from familiar.skills.bookkeeping.skill import get_all_transactions

    txns = get_all_transactions()
    return jsonify({"transaction": txns[0] if txns else None}), 201


@bp.route("/api/bookkeeping/transactions/<tx_id>", methods=["PUT"])
@require_auth
def api_bookkeeping_transaction_update(tx_id):
    """Update a transaction."""
    data = request.json or {}
    data["id"] = tx_id

    from familiar.skills.bookkeeping.skill import update_transaction

    result = update_transaction(data)
    if "error" in result:
        return jsonify(result), 404
    return jsonify({"transaction": result})


@bp.route("/api/bookkeeping/transactions/<tx_id>", methods=["DELETE"])
@require_auth
def api_bookkeeping_transaction_delete(tx_id):
    """Delete a transaction."""
    from familiar.skills.bookkeeping.skill import delete_transaction

    result = delete_transaction(tx_id)
    if "error" in result:
        return jsonify(result), 404
    return jsonify(result)


@bp.route("/api/bookkeeping/categories")
@require_auth
def api_bookkeeping_categories():
    """Get income/expense category lists."""
    from familiar.skills.bookkeeping.skill import get_categories

    return jsonify({"categories": get_categories()})


@bp.route("/api/bookkeeping/funds")
@require_auth
def api_bookkeeping_funds():
    """Get fund list."""
    from familiar.skills.bookkeeping.skill import get_funds

    return jsonify({"funds": get_funds()})


@bp.route("/api/nonprofit/import", methods=["POST"])
@require_auth
def api_nonprofit_import():
    """Import donors from CSV data.

    Expects JSON body: {"rows": [{"name": "...", "email": "...", ...}, ...]}
    or a CSV file upload via multipart form.
    """
    from familiar.skills.nonprofit.skill import log_donor

    rows = []
    if request.is_json:
        rows = (request.json or {}).get("rows", [])
    elif request.files.get("file"):
        import csv
        import io

        f = request.files["file"]
        text = f.read().decode("utf-8-sig")
        reader = csv.DictReader(io.StringIO(text))
        rows = list(reader)

    if not rows:
        return jsonify({"error": "No data provided."}), 400

    imported = 0
    errors = []
    for i, row in enumerate(rows):
        name = (row.get("name") or row.get("Name") or "").strip()
        if not name:
            errors.append(f"Row {i + 1}: missing name")
            continue

        donor_data = {
            "name": name,
            "email": row.get("email") or row.get("Email") or "",
            "phone": row.get("phone") or row.get("Phone") or "",
            "notes": row.get("notes") or row.get("Notes") or "",
        }

        # Handle tags (comma-separated string or list)
        tags = row.get("tags") or row.get("Tags") or ""
        if isinstance(tags, str) and tags:
            donor_data["tags"] = [t.strip() for t in tags.split(",") if t.strip()]

        # If there's an amount, log as gift
        amount = row.get("amount") or row.get("Amount") or row.get("total_given") or ""
        if amount:
            try:
                # Strip currency symbols
                amount_clean = str(amount).replace("$", "").replace(",", "").strip()
                donor_data["amount"] = float(amount_clean)
                donor_data["type"] = "gift"
                donor_data["description"] = row.get("description") or "Imported"
            except (ValueError, TypeError):
                pass

        log_donor(donor_data)
        imported += 1

    return jsonify(
        {
            "imported": imported,
            "errors": errors,
            "total_rows": len(rows),
        }
    )


# ============================================================
# Financial Charts API
# ============================================================


@bp.route("/api/bookkeeping/chart/<chart_type>")
@require_auth
def api_bookkeeping_chart(chart_type):
    """Generate a financial chart as base64 PNG.

    Chart types:
      - monthly: income vs expense bar chart by month (current year)
      - category_income: pie chart of income by category
      - category_expense: pie chart of expense by category
      - fund: bar chart of fund balances
    """
    import base64
    from datetime import datetime as _dt

    from familiar.skills.bookkeeping.skill import get_all_transactions

    year = request.args.get("year", str(_dt.now().year))
    txns = get_all_transactions({"start_date": f"{year}-01-01", "end_date": f"{year}-12-31"})

    try:
        from familiar.skills.charts.skill import render_chart_bytes
    except ImportError:
        return jsonify({"error": "Charts skill not available (matplotlib not installed)."}), 501

    chart_bytes = None

    if chart_type == "monthly":
        # Monthly income vs expenses
        months = {}
        for t in txns:
            m = t.get("date", "")[:7]  # "2026-01"
            if m not in months:
                months[m] = {"income": 0, "expense": 0}
            months[m][t["type"]] += t["amount"]
        sorted_months = sorted(months.keys())
        if not sorted_months:
            return jsonify({"error": "No transactions for this year."})
        labels = [m[5:] for m in sorted_months]  # "01", "02", etc.
        month_names = {
            "01": "Jan",
            "02": "Feb",
            "03": "Mar",
            "04": "Apr",
            "05": "May",
            "06": "Jun",
            "07": "Jul",
            "08": "Aug",
            "09": "Sep",
            "10": "Oct",
            "11": "Nov",
            "12": "Dec",
        }
        labels = [month_names.get(l, l) for l in labels]
        chart_bytes = render_chart_bytes(
            chart_type="bar",
            title=f"Income vs Expenses — {year}",
            labels=labels,
            series=[
                {"name": "Income", "values": [months[m]["income"] for m in sorted_months]},
                {"name": "Expenses", "values": [months[m]["expense"] for m in sorted_months]},
            ],
            y_label="Amount ($)",
            currency=True,
        )

    elif chart_type == "category_income":
        cats = {}
        for t in txns:
            if t["type"] == "income":
                cat = t.get("category", "other").replace("_", " ").title()
                cats[cat] = cats.get(cat, 0) + t["amount"]
        if not cats:
            return jsonify({"error": "No income transactions."})
        labels = list(cats.keys())
        values = list(cats.values())
        chart_bytes = render_chart_bytes(
            chart_type="pie",
            title=f"Income by Category — {year}",
            labels=labels,
            values=values,
            currency=True,
        )

    elif chart_type == "category_expense":
        cats = {}
        for t in txns:
            if t["type"] == "expense":
                cat = t.get("category", "other").replace("_", " ").title()
                cats[cat] = cats.get(cat, 0) + t["amount"]
        if not cats:
            return jsonify({"error": "No expense transactions."})
        labels = list(cats.keys())
        values = list(cats.values())
        chart_bytes = render_chart_bytes(
            chart_type="pie",
            title=f"Expenses by Category — {year}",
            labels=labels,
            values=values,
            currency=True,
        )

    elif chart_type == "fund":
        funds = {}
        for t in txns:
            fund = t.get("fund", "General")
            if fund not in funds:
                funds[fund] = {"income": 0, "expense": 0}
            funds[fund][t["type"]] += t["amount"]
        if not funds:
            return jsonify({"error": "No transactions."})
        labels = list(funds.keys())
        chart_bytes = render_chart_bytes(
            chart_type="bar",
            title=f"Fund Balances — {year}",
            labels=labels,
            series=[
                {"name": "Income", "values": [funds[f]["income"] for f in labels]},
                {"name": "Expenses", "values": [funds[f]["expense"] for f in labels]},
            ],
            y_label="Amount ($)",
            currency=True,
        )

    else:
        return jsonify({"error": f"Unknown chart type: {chart_type}"}), 400

    if not chart_bytes:
        return jsonify({"error": "Chart rendering failed — matplotlib may not be installed."}), 501

    b64 = base64.b64encode(chart_bytes).decode("ascii")
    return jsonify({"image": f"data:image/png;base64,{b64}", "chart_type": chart_type})


# ============================================================
# Grants.gov Search API
# ============================================================


@bp.route("/api/grants-gov/search")
@require_auth
def api_grants_gov_search():
    """Search Grants.gov for federal funding opportunities."""
    keyword = request.args.get("keyword", "") or request.args.get("q", "")
    agency = request.args.get("agency", "")
    category = request.args.get("category", "")
    date_range = request.args.get("dateRange", "")  # posted within last N days
    sort_by = request.args.get("sortBy", "")  # closeDateAsc, closeDateDesc, openDate, etc.
    limit = min(int(request.args.get("limit", "25")), 100)
    offset = int(request.args.get("offset", "0"))

    if not keyword and not agency and not category:
        return jsonify({"error": "Provide keyword, agency, or category"}), 400

    try:
        from familiar.skills._api_utils import api_post
        from familiar.skills.grants_gov.skill import search_grants as _search_grants  # noqa: F401

        body = {
            "paging": {"num": limit, "offset": offset},
            "criteria": {"oppStatuses": "forecasted|posted"},
        }
        if keyword:
            body["criteria"]["keyword"] = keyword
        if agency:
            body["criteria"]["agency"] = agency
        if category:
            body["criteria"]["fundingCategory"] = category
        if date_range:
            body["criteria"]["dateRange"] = date_range
        if sort_by:
            body["criteria"]["sortBy"] = sort_by

        result = api_post(
            "https://api.grants.gov/v1/api/search2",
            json=body,
            headers={"Content-Type": "application/json"},
            service_name="Grants.gov",
        )
        search_data = result.get("data", result)
        opportunities = search_data.get("oppHits", [])
        grants = []
        for opp in opportunities:
            grants.append(
                {
                    "id": opp.get("id", opp.get("number", "")),
                    "title": opp.get("title", "Untitled"),
                    "agency": opp.get("agency", ""),
                    "open_date": opp.get("openDate", ""),
                    "close_date": opp.get("closeDate", ""),
                    "award_floor": opp.get("awardFloor", 0),
                    "award_ceiling": opp.get("awardCeiling", 0),
                    "number": opp.get("number", ""),
                    "status": opp.get("oppStatus", ""),
                }
            )
        total = search_data.get("hitCount", search_data.get("totalCount", len(grants)))
        return jsonify(
            {
                "grants": grants,
                "total": total,
                "offset": offset,
                "limit": limit,
                "has_more": (offset + limit) < total,
            }
        )
    except Exception as e:
        return jsonify({"error": safe_error(e)}), 502


# ============================================================
# Legislation Search + Tracking API
# ============================================================


@bp.route("/api/legislation/search")
@require_auth
def api_legislation_search():
    """Search Congress.gov for federal bills."""
    query = request.args.get("query", "")
    congress = request.args.get("congress", "119")  # Current congress
    sort = request.args.get("sort", "updateDate+desc")
    limit = min(int(request.args.get("limit", "25")), 100)
    offset = int(request.args.get("offset", "0"))

    if not query:
        return jsonify({"error": "Provide a search query"}), 400

    try:
        import urllib.error
        import urllib.request

        api_key = os.environ.get("CONGRESS_API_KEY", "DEMO_KEY")
        params = (
            f"query={urllib.parse.quote(query)}"
            f"&limit={limit}&offset={offset}"
            f"&sort={urllib.parse.quote(sort)}"
            f"&format=json&api_key={api_key}"
        )
        url = f"https://api.congress.gov/v3/bill/{congress}?{params}"

        req = urllib.request.Request(
            url,
            headers={
                "Accept": "application/json",
                "User-Agent": "Familiar-Agent/1.15 (nonprofit-tool)",
            },
        )
        with urllib.request.urlopen(req, timeout=15) as resp:
            data = json.loads(resp.read())

        bills = []
        for b in data.get("bills", []):
            latest = b.get("latestAction", {})
            bills.append(
                {
                    "bill_id": f"{b.get('type', '')}{b.get('number', '')}",
                    "title": b.get("title", ""),
                    "chamber": b.get("originChamber", ""),
                    "congress": str(b.get("congress", congress)),
                    "latest_action": latest.get("text", ""),
                    "latest_action_date": latest.get("actionDate", ""),
                    "url": b.get("url", "")
                    .replace("format=json", "")
                    .replace("api_key=" + api_key, "")
                    .rstrip("?&"),
                    "type": b.get("type", ""),
                    "number": str(b.get("number", "")),
                    "update_date": b.get("updateDate", ""),
                }
            )

        pagination = data.get("pagination", {})
        total = pagination.get("count", len(bills))

        return jsonify(
            {
                "bills": bills,
                "total": total,
                "offset": offset,
                "limit": limit,
                "has_more": (offset + limit) < total,
            }
        )
    except Exception as e:
        return jsonify({"error": safe_error(e)}), 502


@bp.route("/api/legislation/search/state")
@require_auth
def api_legislation_search_state():
    """Search state bills via OpenStates or LegiScan.

    Tries OpenStates first (OPENSTATES_API_KEY), falls back to LegiScan
    (LEGISCAN_API_KEY). Returns 501 if neither key is configured.
    """
    query = request.args.get("query", "")
    state = request.args.get("state", "")
    limit = min(int(request.args.get("limit", "25")), 50)
    page = int(request.args.get("page", "1"))

    if not query:
        return jsonify({"error": "Provide a search query"}), 400
    if not state:
        return jsonify({"error": "Provide a state (e.g. WA, CA, NY)"}), 400

    import urllib.error
    import urllib.request

    # --- Try OpenStates first ---
    os_key = os.environ.get("OPENSTATES_API_KEY", "")
    if os_key:
        try:
            params = (
                f"jurisdiction={urllib.parse.quote(state)}"
                f"&q={urllib.parse.quote(query)}"
                f"&page={page}&per_page={limit}"
                f"&sort=updated_desc"
                f"&apikey={os_key}"
            )
            url = f"https://v3.openstates.org/bills?{params}"
            req = urllib.request.Request(
                url,
                headers={
                    "Accept": "application/json",
                    "User-Agent": "Familiar-Agent/1.15 (nonprofit-tool)",
                },
            )
            with urllib.request.urlopen(req, timeout=15) as resp:
                data = json.loads(resp.read())

            bills = []
            for b in data.get("results", []):
                latest = b.get("latest_action_description") or ""
                latest_date = (b.get("latest_action_date") or "")[:10]
                bills.append(
                    {
                        "bill_id": b.get("identifier", ""),
                        "title": b.get("title", ""),
                        "chamber": b.get("from_organization", {}).get("name", ""),
                        "congress": b.get("session", ""),
                        "state": state.upper(),
                        "level": "state",
                        "latest_action": latest,
                        "latest_action_date": latest_date,
                        "url": b.get("openstates_url", ""),
                        "sponsor": (b["sponsorships"][0]["name"] if b.get("sponsorships") else ""),
                        "update_date": (b.get("updated_at") or "")[:10],
                    }
                )
            pagination = data.get("pagination", {})
            total = pagination.get("total_items", len(bills))
            max_page = pagination.get("max_page", 1)
            return jsonify(
                {
                    "bills": bills,
                    "total": total,
                    "page": page,
                    "limit": limit,
                    "has_more": page < max_page,
                    "source": "openstates",
                }
            )
        except Exception as e:
            logger.warning("OpenStates search failed: %s", e)
            # Fall through to LegiScan

    # --- Try LegiScan ---
    ls_key = os.environ.get("LEGISCAN_API_KEY", "")
    if ls_key:
        try:
            params = (
                f"key={ls_key}&op=searchRaw"
                f"&state={urllib.parse.quote(state.upper())}"
                f"&query={urllib.parse.quote(query)}"
                f"&page={page}"
            )
            url = f"https://api.legiscan.com/?{params}"
            req = urllib.request.Request(
                url,
                headers={
                    "User-Agent": "Familiar-Agent/1.15 (nonprofit-tool)",
                },
            )
            with urllib.request.urlopen(req, timeout=15) as resp:
                data = json.loads(resp.read())

            if data.get("status") == "ERROR":
                return jsonify(
                    {"error": data.get("alert", {}).get("message", "LegiScan error")}
                ), 502

            search_result = data.get("searchresult", {})
            # LegiScan returns {summary: {}, 0: {}, 1: {}, ...}
            summary = search_result.pop("summary", {})
            total = summary.get("count", 0)

            bills = []
            for _key, b in sorted(search_result.items()):
                if not isinstance(b, dict) or "bill_number" not in b:
                    continue
                bills.append(
                    {
                        "bill_id": b.get("bill_number", ""),
                        "title": b.get("title", ""),
                        "chamber": "",
                        "congress": str(b.get("session_id", "")),
                        "state": b.get("state", state.upper()),
                        "level": "state",
                        "latest_action": b.get("last_action", ""),
                        "latest_action_date": b.get("last_action_date", ""),
                        "url": b.get("url", ""),
                        "sponsor": "",
                        "update_date": b.get("last_action_date", ""),
                    }
                )

            return jsonify(
                {
                    "bills": bills,
                    "total": total,
                    "page": page,
                    "limit": limit,
                    "has_more": page * 50 < total,  # LegiScan pages are 50 items
                    "source": "legiscan",
                }
            )
        except Exception as e:
            logger.warning("LegiScan search failed: %s", e)

    # --- Fallback: WA state has a free API (no key needed) ---
    if state.upper() == "WA":
        try:
            import xml.etree.ElementTree as _ET

            _wa_url = (
                "https://wslwebservices.leg.wa.gov/LegislationService.asmx/"
                "GetLegislativeStatusChangesByDateRange"
                "?biennium=2025-26"
                "&beginDate=2025-01-01&endDate=2026-12-31"
            )
            req = urllib.request.Request(
                _wa_url,
                headers={
                    "User-Agent": "Familiar-Agent/1.15 (nonprofit-tool)",
                },
            )
            with urllib.request.urlopen(req, timeout=15) as resp:
                xml_data = resp.read()
            root = _ET.fromstring(xml_data)
            ns = {"ws": "http://WSLWebServices.leg.wa.gov/"}
            changes = root.findall(".//ws:LegislativeStatus", ns)

            # Filter by keyword, deduplicate by bill ID, take most recent action
            q_lower = query.lower()
            bill_map = {}
            for c in changes:
                bid_el = c.find("ws:BillId", ns)
                action_el = c.find("ws:HistoryLine", ns)
                date_el = c.find("ws:ActionDate", ns)
                bid = bid_el.text if bid_el is not None else ""
                action = action_el.text if action_el is not None else ""
                adate = (date_el.text or "")[:10] if date_el is not None else ""

                # Match keyword against bill ID or action text
                if q_lower not in bid.lower() and q_lower not in action.lower():
                    continue

                if bid not in bill_map or adate > bill_map[bid]["latest_action_date"]:
                    bill_map[bid] = {
                        "bill_id": bid,
                        "title": bid,  # WA API doesn't include title in status changes
                        "chamber": "House" if bid.startswith("H") else "Senate",
                        "congress": "2025-26",
                        "state": "WA",
                        "level": "state",
                        "latest_action": action,
                        "latest_action_date": adate,
                        "url": f"https://app.leg.wa.gov/billsummary?BillNumber="
                        f"{bid.replace('HB ', '').replace('SB ', '').replace('SHB ', '').replace('SSB ', '').split()[0]}"
                        f"&Year=2025",
                        "sponsor": "",
                        "update_date": adate,
                    }

            bills = sorted(bill_map.values(), key=lambda b: b["latest_action_date"], reverse=True)
            total = len(bills)
            start = (page - 1) * limit
            page_bills = bills[start : start + limit]

            return jsonify(
                {
                    "bills": page_bills,
                    "total": total,
                    "page": page,
                    "limit": limit,
                    "has_more": start + limit < total,
                    "source": "WA Legislature (leg.wa.gov)",
                }
            )
        except Exception as e:
            logger.warning("WA Legislature search failed: %s", e)

    # Neither configured and no built-in fallback for this state
    return jsonify(
        {
            "error": (
                "State legislation search requires an API key for most states. "
                "Set OPENSTATES_API_KEY (free at openstates.org/accounts/signup) "
                "or LEGISCAN_API_KEY (free at legiscan.com/legiscan) "
                "in Settings > Credentials. "
                "Washington State (WA) and Seattle work without keys."
            ),
            "setup_hint": True,
        }
    ), 501


@bp.route("/api/legislation/search/city")
@require_auth
def api_legislation_search_city():
    """Search city council legislation via Legistar API.

    Currently supports Seattle. Extensible to any city using Legistar
    (NYC, Chicago, LA, Portland, etc.) by adding to LEGISTAR_CITIES.
    """
    query = request.args.get("query", "")
    city = request.args.get("city", "seattle").lower()
    limit = min(int(request.args.get("limit", "25")), 50)
    page = int(request.args.get("page", "1"))

    if not query:
        return jsonify({"error": "Provide a search query"}), 400

    # Cities that use Legistar (free, no key needed)
    LEGISTAR_CITIES = {
        "seattle": {"client": "seattle", "label": "Seattle City Council"},
        "nyc": {"client": "nyc", "label": "New York City Council"},
        "chicago": {"client": "chicago", "label": "Chicago City Council"},
        "la": {"client": "lacity", "label": "Los Angeles City Council"},
        "portland": {"client": "portland", "label": "Portland City Council"},
        "denver": {"client": "denver", "label": "Denver City Council"},
        "philadelphia": {"client": "phlapi", "label": "Philadelphia City Council"},
        "san_francisco": {"client": "sfgov", "label": "San Francisco Board of Supervisors"},
    }

    city_cfg = LEGISTAR_CITIES.get(city)
    if not city_cfg:
        return jsonify(
            {
                "error": f"City '{city}' not supported yet. Available: {', '.join(LEGISTAR_CITIES.keys())}",
                "available_cities": list(LEGISTAR_CITIES.keys()),
            }
        ), 400

    try:
        import urllib.parse
        import urllib.request

        skip = (page - 1) * limit
        # Legistar OData filter — search title text
        odata_filter = urllib.parse.quote(f"substringof('{query}', MatterTitle) eq true")
        url = (
            f"https://webapi.legistar.com/v1/{city_cfg['client']}/matters"
            f"?$top={limit}&$skip={skip}"
            f"&$filter={odata_filter}"
            f"&$orderby=MatterLastModifiedUtc+desc"
        )
        req = urllib.request.Request(
            url,
            headers={
                "Accept": "application/json",
                "User-Agent": "Familiar-Agent/1.15 (nonprofit-tool)",
            },
        )
        with urllib.request.urlopen(req, timeout=15) as resp:
            matters = json.loads(resp.read())

        bills = []
        for m in matters or []:
            title = m.get("MatterTitle", "") or m.get("MatterName", "")
            bills.append(
                {
                    "bill_id": m.get("MatterFile", ""),
                    "title": title[:200],
                    "chamber": m.get("MatterBodyName", ""),
                    "congress": "",
                    "state": "",
                    "level": "city",
                    "city": city,
                    "latest_action": m.get("MatterStatusName", ""),
                    "latest_action_date": (m.get("MatterIntroDate") or "")[:10],
                    "url": (
                        f"https://{city_cfg['client']}.legistar.com/"
                        f"LegislationDetail.aspx?ID={m.get('MatterId', '')}"
                        f"&GUID={m.get('MatterGuid', '')}"
                    ),
                    "sponsor": "",
                    "type": m.get("MatterTypeName", ""),
                    "update_date": (m.get("MatterLastModifiedUtc") or "")[:10],
                }
            )

        # Legistar doesn't return total count easily — estimate from results
        has_more = len(bills) == limit
        total = skip + len(bills) + (1 if has_more else 0)

        return jsonify(
            {
                "bills": bills,
                "total": total,
                "page": page,
                "limit": limit,
                "has_more": has_more,
                "source": city_cfg["label"] + " (Legistar)",
            }
        )
    except Exception as e:
        return jsonify({"error": safe_error(e)}), 502


@bp.route("/api/nonprofit/legislation")
@require_auth
def api_legislation_list():
    """List all tracked legislation (uses advocacy BillStore)."""
    try:
        from familiar.skills.advocacy.skill import get_bill_store

        store = get_bill_store()
        return jsonify({"legislation": store.list_bills()})
    except Exception as e:
        return jsonify({"legislation": [], "error": safe_error(e)})


@bp.route("/api/nonprofit/legislation", methods=["POST"])
@require_auth
def api_legislation_add():
    """Track a bill via advocacy BillStore."""
    data = request.json or {}
    if not data.get("title", "").strip():
        return jsonify({"error": "Bill title is required."}), 400
    try:
        from familiar.skills.advocacy.skill import get_bill_store

        store = get_bill_store()
        bill_id = store.track_bill(
            bill_number=data.get("bill_id", ""),
            title=data.get("title", ""),
            jurisdiction="federal" if data.get("level") == "federal" else data.get("state", ""),
            session=data.get("congress", ""),
            status="introduced",
            our_position="watch",
            priority=data.get("priority", "medium"),
            sponsor=data.get("sponsor", ""),
            summary=data.get("latest_action", ""),
        )
        if bill_id is None:
            return jsonify({"error": "Bill already tracked."}), 409
        bill = store.get_bill(bill_id)
        return jsonify({"bill": bill}), 201
    except Exception as e:
        return jsonify({"error": safe_error(e)}), 500


@bp.route("/api/nonprofit/legislation/<bill_id>", methods=["DELETE"])
@require_auth
def api_legislation_delete(bill_id):
    """Stop tracking a bill."""
    try:
        from familiar.skills.advocacy.skill import get_bill_store

        store = get_bill_store()
        ok = store.delete_bill(bill_id)
        return jsonify({"deleted": ok}), 200 if ok else 404
    except Exception as e:
        return jsonify({"error": safe_error(e)}), 500


@bp.route("/api/nonprofit/watchlist")
@require_auth
def api_watchlist_list():
    """Get bill keyword watchlist."""
    from familiar.skills.nonprofit.skill import get_bill_watchlist

    return jsonify({"watchlist": get_bill_watchlist()})


@bp.route("/api/nonprofit/watchlist", methods=["POST"])
@require_auth
def api_watchlist_add():
    """Add keyword to watchlist."""
    data = request.json or {}
    if not data.get("keyword", "").strip():
        return jsonify({"error": "Keyword is required."}), 400
    from familiar.skills.nonprofit.skill import add_watchlist_keyword

    msg = add_watchlist_keyword(data)
    from familiar.skills.nonprofit.skill import get_bill_watchlist

    return jsonify({"message": msg, "watchlist": get_bill_watchlist()}), 201


@bp.route("/api/nonprofit/watchlist/<keyword_id>", methods=["DELETE"])
@require_auth
def api_watchlist_delete(keyword_id):
    """Remove keyword from watchlist."""
    from familiar.skills.nonprofit.skill import remove_watchlist_keyword

    ok = remove_watchlist_keyword(keyword_id)
    return jsonify({"deleted": ok}), 200 if ok else 404



# ============================================================
# CSV Export Endpoints
# ============================================================


@bp.route("/api/nonprofit/donors/export")
@require_auth
def api_nonprofit_donors_export():
    """Export donors as CSV file download."""
    import csv
    import io

    from familiar.skills.nonprofit.skill import get_all_donors

    donors = get_all_donors()
    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerow(
        [
            "name",
            "email",
            "phone",
            "total_given",
            "first_gift",
            "last_gift",
            "tags",
            "notes",
            "interactions_count",
        ]
    )
    for d in donors:
        writer.writerow(
            [
                d.get("name", ""),
                d.get("email", ""),
                d.get("phone", ""),
                d.get("total_given", 0),
                (d.get("first_gift") or "")[:10],
                (d.get("last_gift") or "")[:10],
                ", ".join(d.get("tags", [])),
                d.get("notes", ""),
                len(d.get("interactions", [])),
            ]
        )

    resp = make_response(output.getvalue())
    resp.headers["Content-Disposition"] = "attachment; filename=donors.csv"
    resp.headers["Content-Type"] = "text/csv; charset=utf-8"
    return resp


@bp.route("/api/nonprofit/grants/export")
@require_auth
def api_nonprofit_grants_export():
    """Export grants as CSV file download."""
    import csv
    import io

    from familiar.skills.nonprofit.skill import get_all_grants

    grants = get_all_grants()
    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerow(
        [
            "name",
            "funder",
            "amount",
            "status",
            "deadline",
            "report_due",
            "start_date",
            "end_date",
            "notes",
        ]
    )
    for g in grants:
        writer.writerow(
            [
                g.get("name", ""),
                g.get("funder", ""),
                g.get("amount", ""),
                g.get("status", ""),
                g.get("deadline", ""),
                g.get("report_due", ""),
                g.get("start_date", ""),
                g.get("end_date", ""),
                g.get("notes", ""),
            ]
        )

    resp = make_response(output.getvalue())
    resp.headers["Content-Disposition"] = "attachment; filename=grants.csv"
    resp.headers["Content-Type"] = "text/csv; charset=utf-8"
    return resp


@bp.route("/api/bookkeeping/transactions/export")
@require_auth
def api_bookkeeping_transactions_export():
    """Export transactions as CSV file download."""
    import csv
    import io

    from familiar.skills.bookkeeping.skill import get_all_transactions

    filters = {
        "start_date": request.args.get("start_date"),
        "end_date": request.args.get("end_date"),
        "category": request.args.get("category"),
        "fund": request.args.get("fund"),
        "type": request.args.get("type"),
    }
    filters = {k: v for k, v in filters.items() if v}
    txns = get_all_transactions(filters if filters else None)

    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerow(
        ["date", "type", "amount", "category", "fund", "source", "vendor", "description", "id"]
    )
    for t in txns:
        writer.writerow(
            [
                t.get("date", ""),
                t.get("type", ""),
                t.get("amount", ""),
                t.get("category", ""),
                t.get("fund", ""),
                t.get("source", ""),
                t.get("vendor", ""),
                t.get("description", ""),
                t.get("id", ""),
            ]
        )

    resp = make_response(output.getvalue())
    resp.headers["Content-Disposition"] = "attachment; filename=transactions.csv"
    resp.headers["Content-Type"] = "text/csv; charset=utf-8"
    return resp


# ============================================================
# Document Generation Endpoints
# ============================================================


@bp.route("/api/nonprofit/generate/receipt", methods=["POST"])
@require_auth
def api_nonprofit_generate_receipt():
    """Generate a tax-deductible donation receipt for a donor."""
    from familiar.skills.documents.skill import create_receipt

    data = request.get_json(force=True)
    donor_name = data.get("donor_name", "Donor")
    amount = data.get("amount", 0)
    date_of_gift = data.get("date", "")
    description = data.get("description", "Charitable donation")

    result = create_receipt(
        {
            "donor_name": donor_name,
            "amount": amount,
            "date_of_gift": date_of_gift,
            "description": description,
            "format": "docx",
        }
    )
    # result is a string like "Receipt saved to /path/to/file.docx"
    # Extract path and send as download
    return _send_generated_doc(result)


@bp.route("/api/nonprofit/generate/thankyou", methods=["POST"])
@require_auth
def api_nonprofit_generate_thankyou():
    """Generate a thank-you letter for a donor using the bundled template."""
    from familiar.skills.documents.skill import fill_template

    data = request.get_json(force=True)
    fields = {
        "donor_name": data.get("donor_name", "Valued Donor"),
        "amount": str(data.get("amount", "")),
        "organization": data.get("organization", "Our Organization"),
        "gift_date": data.get("gift_date", ""),
        "impact_area": data.get("impact_area", "our mission"),
        "personal_note": data.get("personal_note", ""),
        "mission_statement": data.get("mission_statement", ""),
        "signer_name": data.get("signer_name", ""),
        "signer_title": data.get("signer_title", "Executive Director"),
    }

    result = fill_template(
        {
            "template": "nonprofit_director_donor_thankyou",
            "fields": fields,
            "format": "docx",
        }
    )
    return _send_generated_doc(result)


@bp.route("/api/nonprofit/generate/board-report", methods=["POST"])
@require_auth
def api_nonprofit_generate_board_report():
    """Generate a board report with auto-populated financial data."""
    from datetime import datetime as _dt

    from familiar.skills.bookkeeping.skill import get_all_transactions
    from familiar.skills.documents.skill import fill_template
    from familiar.skills.nonprofit.skill import get_all_donors, get_all_grants

    now = _dt.now()
    year = str(now.year)
    txns = get_all_transactions({"start_date": f"{year}-01-01", "end_date": f"{year}-12-31"})

    # Compute financials
    total_income = sum(t["amount"] for t in txns if t["type"] == "income")
    total_expense = sum(t["amount"] for t in txns if t["type"] == "expense")
    net = total_income - total_expense

    donors = get_all_donors()
    grants = get_all_grants()
    active_grants = [g for g in grants if g.get("status") in ("applied", "awarded")]

    # Build sections text
    fin_overview = (
        f"Year-to-date income: ${total_income:,.2f}\n"
        f"Year-to-date expenses: ${total_expense:,.2f}\n"
        f"Net position: ${net:,.2f}\n"
        f"Total transactions: {len(txns)}"
    )

    fundraising = (
        f"Total donors: {len(donors)}\n"
        f"Total donated (all time): ${sum(d.get('total_given', 0) for d in donors):,.2f}"
    )

    grant_lines = []
    for g in active_grants:
        status = g.get("status", "unknown").title()
        amt = g.get("amount", 0)
        grant_lines.append(f"- {g.get('name', 'Unnamed')}: {status} (${amt:,.0f})")
    grant_text = "\n".join(grant_lines) if grant_lines else "No active grants."

    data = request.get_json(force=True) or {}
    report_period = data.get("report_period", now.strftime("%B %Y"))

    fields = {
        "report_period": report_period,
        "executive_summary": data.get(
            "executive_summary",
            f"Financial and program summary for {report_period}.",
        ),
        "financial_overview": fin_overview,
        "program_updates": data.get("program_updates", "See grant pipeline below."),
        "fundraising_report": fundraising + "\n\nGrant Pipeline:\n" + grant_text,
        "upcoming_events": data.get("upcoming_events", "None scheduled."),
        "action_items": data.get("action_items", "None."),
        "motions": data.get("motions", "None."),
    }

    result = fill_template(
        {
            "template": "nonprofit_director_board_report",
            "fields": fields,
            "format": "docx",
        }
    )
    return _send_generated_doc(result)


@bp.route("/api/nonprofit/generate/990-data", methods=["POST"])
@require_auth
def api_nonprofit_generate_990_data():
    """Generate 990 filing data export as a structured report."""
    from datetime import datetime as _dt

    from familiar.skills.bookkeeping.skill import get_all_transactions
    from familiar.skills.documents.skill import create_report
    from familiar.skills.nonprofit.skill import get_all_donors, get_all_grants

    data = request.get_json(force=True) or {}
    year = data.get("year", str(_dt.now().year))
    txns = get_all_transactions({"start_date": f"{year}-01-01", "end_date": f"{year}-12-31"})

    income_txns = [t for t in txns if t["type"] == "income"]
    expense_txns = [t for t in txns if t["type"] == "expense"]
    total_income = sum(t["amount"] for t in income_txns)
    total_expense = sum(t["amount"] for t in expense_txns)

    # Income by category
    income_cats = {}
    for t in income_txns:
        cat = t.get("category", "other").replace("_", " ").title()
        income_cats[cat] = income_cats.get(cat, 0) + t["amount"]

    # Expense by category
    expense_cats = {}
    for t in expense_txns:
        cat = t.get("category", "other").replace("_", " ").title()
        expense_cats[cat] = expense_cats.get(cat, 0) + t["amount"]

    donors = get_all_donors()
    grants = get_all_grants()
    awarded = [g for g in grants if g.get("status") == "awarded"]

    sections = [
        {
            "heading": "Revenue Summary",
            "body": f"Total Revenue: ${total_income:,.2f}",
            "table": {
                "headers": ["Category", "Amount"],
                "rows": [
                    [k, f"${v:,.2f}"]
                    for k, v in sorted(income_cats.items(), key=lambda x: x[1], reverse=True)
                ],
            },
        },
        {
            "heading": "Expense Summary",
            "body": f"Total Expenses: ${total_expense:,.2f}",
            "table": {
                "headers": ["Category", "Amount"],
                "rows": [
                    [k, f"${v:,.2f}"]
                    for k, v in sorted(expense_cats.items(), key=lambda x: x[1], reverse=True)
                ],
            },
        },
        {
            "heading": "Net Assets",
            "body": (
                f"Net Income: ${total_income - total_expense:,.2f}\nTotal transactions: {len(txns)}"
            ),
        },
        {
            "heading": "Contributor Information",
            "body": (
                f"Total individual contributors: {len(donors)}\n"
                f"Total individual contributions: "
                f"${sum(d.get('total_given', 0) for d in donors):,.2f}\n"
                f"Active grants received: {len(awarded)}\n"
                f"Grant revenue: "
                f"${sum(g.get('amount', 0) for g in awarded):,.2f}"
            ),
        },
    ]

    result = create_report(
        {
            "title": f"Form 990 Data — Fiscal Year {year}",
            "sections": sections,
            "format": "docx",
            "letterhead": True,
        }
    )
    return _send_generated_doc(result)


# ============================================================
# Nonprofit Email Pipeline API — Pending Events + Action Queue
# ============================================================


@bp.route("/api/nonprofit/pending-events")
@require_auth
def api_nonprofit_pending_events():
    """List all pending calendar events detected from emails."""
    from familiar.skills.email.pending_events import get_pending_event_store

    store = get_pending_event_store()
    return jsonify({"events": store.get_pending(), "count": store.count_pending()})


@bp.route("/api/nonprofit/pending-events/<event_id>/confirm", methods=["POST"])
@require_auth
def api_nonprofit_pending_event_confirm(event_id):
    """Confirm a pending event (add to calendar)."""
    from familiar.skills.email.pending_events import get_pending_event_store

    store = get_pending_event_store()
    if store.mark_confirmed(event_id):
        return jsonify({"status": "confirmed", "id": event_id})
    return jsonify({"error": "Event not found."}), 404


@bp.route("/api/nonprofit/pending-events/<event_id>/dismiss", methods=["POST"])
@require_auth
def api_nonprofit_pending_event_dismiss(event_id):
    """Dismiss a pending event (don't add to calendar)."""
    from familiar.skills.email.pending_events import get_pending_event_store

    store = get_pending_event_store()
    if store.mark_dismissed(event_id):
        return jsonify({"status": "dismissed", "id": event_id})
    return jsonify({"error": "Event not found."}), 404


@bp.route("/api/nonprofit/action-queue")
@require_auth
def api_nonprofit_action_queue():
    """List all pending email follow-up actions."""
    from familiar.skills.email.action_queue import get_action_store

    store = get_action_store()
    return jsonify(
        {
            "actions": store.get_pending(),
            "count": store.count_pending(),
            "by_type": store.summary_by_type(),
        }
    )


@bp.route("/api/nonprofit/action-queue/<action_id>/complete", methods=["POST"])
@require_auth
def api_nonprofit_action_complete(action_id):
    """Mark an action as completed."""
    from familiar.skills.email.action_queue import get_action_store

    store = get_action_store()
    if store.mark_complete(action_id):
        return jsonify({"status": "completed", "id": action_id})
    return jsonify({"error": "Action not found."}), 404


@bp.route("/api/nonprofit/action-queue/<action_id>/delegate", methods=["POST"])
@require_auth
def api_nonprofit_action_delegate(action_id):
    """Delegate an action to a staff member."""
    data = request.json or {}
    delegate_to = data.get("delegate_to", "").strip()
    if not delegate_to:
        return jsonify({"error": "delegate_to is required."}), 400

    from familiar.skills.email.action_queue import get_action_store

    store = get_action_store()
    if store.mark_delegated(action_id, delegate_to):
        return jsonify({"status": "delegated", "id": action_id, "delegate_to": delegate_to})
    return jsonify({"error": "Action not found."}), 404


# ============================================================
# Event Management API
# ============================================================


@bp.route("/api/nonprofit/events")
@require_auth
def api_nonprofit_events():
    """List all events."""
    from familiar.skills.events.skill import get_event_store

    store = get_event_store()
    status = request.args.get("status")
    return jsonify({"events": store.list_events(status=status)})


@bp.route("/api/nonprofit/events", methods=["POST"])
@require_auth
def api_nonprofit_event_create():
    """Create a new event."""
    data = request.json or {}
    if not data.get("name", "").strip():
        return jsonify({"error": "Event name is required."}), 400
    from familiar.skills.events.skill import get_event_store

    store = get_event_store()
    eid = store.create_event(
        name=data["name"].strip(),
        event_type=data.get("event_type", "meeting"),
        start_date=data.get("start_date", ""),
        end_date=data.get("end_date", ""),
        budget_total=float(data.get("budget_total", 0)),
        description=data.get("description", ""),
        volunteer_slots_total=int(data.get("volunteer_slots_total", 0)),
        venue=data.get("venue"),
        partner_orgs=data.get("partner_orgs"),
    )
    return jsonify({"event": store.get_event(eid)}), 201


@bp.route("/api/nonprofit/events/<event_id>", methods=["PUT"])
@require_auth
def api_nonprofit_event_update(event_id):
    """Update an event."""
    data = request.json or {}
    from familiar.skills.events.skill import get_event_store

    store = get_event_store()
    result = store.update_event(event_id, data)
    if result:
        return jsonify({"event": result})
    return jsonify({"error": "Event not found."}), 404


@bp.route("/api/nonprofit/events/<event_id>", methods=["DELETE"])
@require_auth
def api_nonprofit_event_delete(event_id):
    """Delete an event."""
    from familiar.skills.events.skill import get_event_store

    store = get_event_store()
    if store.delete_event(event_id):
        return jsonify({"status": "deleted"})
    return jsonify({"error": "Event not found."}), 404


@bp.route("/api/nonprofit/events/<event_id>/vendors")
@require_auth
def api_nonprofit_event_vendors(event_id):
    """List vendors for an event."""
    from familiar.skills.events.skill import get_vendor_store

    return jsonify({"vendors": get_vendor_store().list_vendors(event_id=event_id)})


@bp.route("/api/nonprofit/events/<event_id>/vendors", methods=["POST"])
@require_auth
def api_nonprofit_event_vendor_add(event_id):
    """Add a vendor to an event."""
    data = request.json or {}
    if not data.get("name", "").strip():
        return jsonify({"error": "Vendor name is required."}), 400
    from familiar.skills.events.skill import get_vendor_store

    store = get_vendor_store()
    vid = store.add_vendor(
        event_id=event_id,
        name=data["name"].strip(),
        service_type=data.get("service_type", "other"),
        contact_name=data.get("contact_name", ""),
        contact_email=data.get("contact_email", ""),
        quote_amount=float(data.get("quote_amount", 0)),
    )
    return jsonify({"vendor_id": vid}), 201


@bp.route("/api/nonprofit/events/<event_id>/volunteers")
@require_auth
def api_nonprofit_event_volunteers(event_id):
    """List volunteers for an event."""
    from familiar.skills.events.skill import get_volunteer_store

    return jsonify({"volunteers": get_volunteer_store().list_volunteers(event_id=event_id)})


@bp.route("/api/nonprofit/events/<event_id>/volunteers", methods=["POST"])
@require_auth
def api_nonprofit_event_volunteer_add(event_id):
    """Add a volunteer to an event."""
    data = request.json or {}
    if not data.get("name", "").strip():
        return jsonify({"error": "Volunteer name is required."}), 400
    from familiar.skills.events.skill import get_volunteer_store

    store = get_volunteer_store()
    vid = store.add_volunteer(
        event_id=event_id,
        name=data["name"].strip(),
        email=data.get("email", ""),
        role=data.get("role", "other"),
        shift_start=data.get("shift_start", ""),
        shift_end=data.get("shift_end", ""),
    )
    return jsonify({"volunteer_id": vid}), 201


# ============================================================
# Advocacy & Legislative API
# ============================================================


@bp.route("/api/nonprofit/bills")
@require_auth
def api_nonprofit_bills():
    """List tracked bills."""
    from familiar.skills.advocacy.skill import get_bill_store

    store = get_bill_store()
    return jsonify(
        {
            "bills": store.list_bills(
                position=request.args.get("position"),
                priority=request.args.get("priority"),
            )
        }
    )


@bp.route("/api/nonprofit/bills", methods=["POST"])
@require_auth
def api_nonprofit_bill_add():
    """Track a new bill."""
    data = request.json or {}
    if not data.get("bill_number", "").strip() or not data.get("title", "").strip():
        return jsonify({"error": "bill_number and title are required."}), 400
    from familiar.skills.advocacy.skill import get_bill_store

    store = get_bill_store()
    bid = store.track_bill(
        bill_number=data["bill_number"].strip(),
        title=data["title"].strip(),
        jurisdiction=data.get("jurisdiction", "WA"),
        status=data.get("status", "introduced"),
        our_position=data.get("our_position", "watch"),
        priority=data.get("priority", "medium"),
        sponsor=data.get("sponsor", ""),
        committee=data.get("committee", ""),
        summary=data.get("summary", ""),
    )
    if bid:
        return jsonify({"bill": store.get_bill(bid)}), 201
    return jsonify({"error": "Bill already tracked."}), 409


@bp.route("/api/nonprofit/bills/<bill_id>", methods=["PUT"])
@require_auth
def api_nonprofit_bill_update(bill_id):
    """Update a tracked bill."""
    data = request.json or {}
    from familiar.skills.advocacy.skill import get_bill_store

    store = get_bill_store()
    result = store.update_bill(bill_id, data)
    if result:
        return jsonify({"bill": result})
    return jsonify({"error": "Bill not found."}), 404


@bp.route("/api/nonprofit/bills/<bill_id>", methods=["DELETE"])
@require_auth
def api_nonprofit_bill_delete(bill_id):
    """Stop tracking a bill."""
    from familiar.skills.advocacy.skill import get_bill_store

    if get_bill_store().remove_bill(bill_id):
        return jsonify({"status": "removed"})
    return jsonify({"error": "Bill not found."}), 404


@bp.route("/api/nonprofit/hearings")
@require_auth
def api_nonprofit_hearings():
    """List upcoming hearings across all bills."""
    days = int(request.args.get("days", 30))
    from familiar.skills.advocacy.skill import get_bill_store

    return jsonify({"hearings": get_bill_store().list_upcoming_hearings(days=days)})


@bp.route("/api/nonprofit/bills/<bill_id>/testimony", methods=["POST"])
@require_auth
def api_nonprofit_bill_testimony(bill_id):
    """Record testimony for a bill."""
    data = request.json or {}
    if not data.get("testifier", "").strip():
        return jsonify({"error": "testifier is required."}), 400
    from familiar.skills.advocacy.skill import get_bill_store

    store = get_bill_store()
    if store.add_testimony(
        bill_id,
        data["testifier"].strip(),
        position=data.get("position", ""),
        format_type=data.get("format", "written"),
    ):
        return jsonify({"status": "recorded"}), 201
    return jsonify({"error": "Bill not found."}), 404


# ============================================================
# Stewardship API
# ============================================================


@bp.route("/api/nonprofit/stewardship")
@require_auth
def api_nonprofit_stewardship():
    """Get stewardship queue — overdue + pending."""
    from familiar.skills.nonprofit.stewardship import get_stewardship_store

    store = get_stewardship_store()
    return jsonify(
        {
            "overdue": store.get_overdue(),
            "pending": store.get_pending(),
            "completion_rate": store.completion_rate(),
        }
    )


@bp.route("/api/nonprofit/stewardship/<entry_id>/complete", methods=["POST"])
@require_auth
def api_nonprofit_stewardship_complete(entry_id):
    """Complete a stewardship action."""
    data = request.json or {}
    index = int(data.get("action_index", 0))
    from familiar.skills.nonprofit.stewardship import get_stewardship_store

    store = get_stewardship_store()
    if store.complete_action(entry_id, index):
        return jsonify({"status": "completed"})
    return jsonify({"error": "Entry or action not found."}), 404


# ============================================================
# Batch Communications API
# ============================================================


@bp.route("/api/nonprofit/batches")
@require_auth
def api_nonprofit_batches():
    """List batch communications."""
    from familiar.skills.nonprofit.batch_comms import get_batch_store

    return jsonify({"batches": get_batch_store().list_batches()})


@bp.route("/api/nonprofit/batches", methods=["POST"])
@require_auth
def api_nonprofit_batch_create():
    """Create a new batch communication."""
    data = request.json or {}
    if not data.get("name", "").strip() and not data.get("segment_key"):
        return jsonify({"error": "name or segment_key is required."}), 400
    from familiar.skills.nonprofit.batch_comms import get_batch_store

    store = get_batch_store()
    bid = store.create_batch(
        name=data.get("name", "").strip(),
        segment_key=data.get("segment_key", ""),
        segment_criteria=data.get("segment_criteria"),
        template_name=data.get("template_name", ""),
        channel=data.get("channel", "email"),
    )
    return jsonify({"batch": store.get_batch(bid)}), 201


@bp.route("/api/nonprofit/batches/<batch_id>")
@require_auth
def api_nonprofit_batch_details(batch_id):
    """Get batch details."""
    from familiar.skills.nonprofit.batch_comms import get_batch_store

    batch = get_batch_store().get_batch(batch_id)
    if batch:
        return jsonify({"batch": batch})
    return jsonify({"error": "Batch not found."}), 404


def _send_generated_doc(result_str):
    """Parse a document skill result string and send the file as download."""
    import re

    # Skill functions return strings like "Receipt saved to /path/to/file.docx"
    # or "Document saved to /path/to/file.docx"
    match = re.search(
        r"(?:saved to|created at|generated at|wrote)\s+(.+\.(?:docx|pdf))",
        result_str,
        re.IGNORECASE,
    )
    if match:
        filepath = match.group(1).strip()
        if os.path.isfile(filepath):
            return send_file(
                filepath,
                as_attachment=True,
                download_name=os.path.basename(filepath),
            )

    # If result looks like an error or no file found, return the raw message
    return jsonify({"message": result_str, "error": "Could not locate generated file."}), 500





"""
Unit tests for Agent — the main coordinator class.

Tests __init__ ordering, backward-compat delegates, chat/chat_stream/chat_with_results
entry points, _setup_request pre-flight, _run_agent_loop phase boundaries,
_finalize_response post-processing, _sync_user_to_session role mapping,
AgentProxy sandboxing, and helper methods.

All 5 managers (ToolExecutor, ProviderManager, PromptBuilder, MemoryManager,
StatusReporter) are mocked — their internals have dedicated test files.
"""

import os
import time
from datetime import datetime, timezone
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from familiar.core.agent import Agent, AgentProxy, _GuardrailBlock, create_agent
from familiar.core.config import load_config
from familiar.core.exceptions import BudgetExceededError
from familiar.core.guardrails import GuardrailViolation, ViolationType
from familiar.core.providers import LLMResponse, StreamEvent, StreamEventType
from familiar.core.security import (
    Capability,
    SecureSession,
    SecurityMode,
    TrustLevel,
)

# ════════════════════════════════════════════════════════════════
# FIXTURES
# ════════════════════════════════════════════════════════════════


@pytest.fixture(autouse=True)
def isolate_filesystem(tmp_path, monkeypatch):
    """Redirect all filesystem side effects to temp dirs."""
    familiar_dir = tmp_path / ".familiar"
    familiar_dir.mkdir()
    (familiar_dir / "episodic").mkdir()
    (familiar_dir / "experiments").mkdir()
    (familiar_dir / "data").mkdir()
    (familiar_dir / "sessions").mkdir()
    monkeypatch.setattr("familiar.core.security.SESSIONS_DIR", tmp_path / ".familiar" / "sessions")
    monkeypatch.setenv("HOME", str(tmp_path))
    return familiar_dir


@pytest.fixture
def mock_provider():
    """A mock LLM provider with sensible defaults."""
    p = MagicMock()
    p.name = "MockProvider"
    p.provider_key = "mock"
    p.model_name = "mock-model"
    p.supports_tools = True
    p.supports_streaming = True
    p.chat.return_value = LLMResponse(
        text="Test response",
        tool_calls=[],
        stop_reason="end_turn",
        usage={"input_tokens": 10, "output_tokens": 8},
    )
    return p


@pytest.fixture
def minimal_config():
    """Config that disables heavy subsystems for fast Agent init."""
    cfg = load_config()
    cfg.agent.memory_enabled = False
    cfg.agent.scheduler_enabled = False
    cfg.agent.planning_enabled = False
    cfg.agent.memory_extraction_enabled = False
    cfg.agent.skills_enabled = False
    cfg.agent.mesh_enabled = False
    cfg.agent.eval_capture_enabled = False
    cfg.agent.distributed_bus_enabled = False
    cfg.llm.default_provider = "ollama"
    cfg.llm.ollama_model = "mock"
    return cfg


def _agent_patches(mock_provider):
    """Context manager stack for creating an Agent with mocked subsystems."""
    return (
        patch("familiar.core.provider_manager.get_provider", return_value=mock_provider),
        patch(
            "familiar.core.agent.get_tool_registry",
            return_value=MagicMock(
                tools={"test_tool": MagicMock()},
            ),
        ),
        patch("familiar.core.agent.SkillLoader", return_value=MagicMock(skills={})),
        patch("familiar.core.agent.get_metrics_collector", return_value=MagicMock()),
        patch("familiar.core.agent.get_tracer", return_value=MagicMock(exporter=None)),
        patch("familiar.core.security.bootstrap_security"),
        patch("familiar.core.agent.is_encryption_enabled", return_value=False),
        patch.dict(os.environ, {"FAMILIAR_ENCRYPTION_KEY": ""}, clear=False),
    )


@pytest.fixture
def agent(minimal_config, mock_provider, tmp_path):
    """Create a real Agent with mocked provider and minimal config."""
    patches = _agent_patches(mock_provider)
    # Enter all patches
    [p.__enter__() if hasattr(p, "__enter__") else p.start() for p in patches]
    try:
        a = Agent(config=minimal_config)
        a.provider = mock_provider
        yield a
    finally:
        for p in reversed(patches):
            try:
                if hasattr(p, "__exit__"):
                    p.__exit__(None, None, None)
                else:
                    p.stop()
            except Exception:
                pass


# ════════════════════════════════════════════════════════════════
# __init__ ORDERING & COMPONENT WIRING
# ════════════════════════════════════════════════════════════════


class TestAgentInit:
    """Agent.__init__ creates all managers in correct order."""

    def test_creates_provider_manager(self, agent):
        assert agent._provider_manager is not None

    def test_creates_tool_executor(self, agent):
        assert agent._tool_executor is not None

    def test_creates_memory_manager(self, agent):
        assert agent._memory_manager is not None

    def test_creates_prompt_builder(self, agent):
        assert agent._prompt_builder is not None

    def test_creates_status_reporter(self, agent):
        assert agent._status_reporter is not None

    def test_creates_guardrails(self, agent):
        assert agent.guardrails is not None

    def test_creates_bus(self, agent):
        assert agent.bus is not None

    def test_creates_health_checker(self, agent):
        assert agent.health is not None

    def test_creates_backup_manager(self, agent):
        assert agent.backup is not None

    def test_creates_sagas(self, agent):
        assert agent.sagas is not None

    def test_creates_experiments(self, agent):
        assert agent.experiments is not None

    def test_creates_orchestrator(self, agent):
        assert agent.orchestrator is not None

    def test_creates_tool_router(self, agent):
        assert agent.tool_router is not None

    def test_creates_retriever(self, agent):
        assert agent.retriever is not None

    def test_provider_set(self, agent, mock_provider):
        assert agent.provider is mock_provider

    def test_config_stored(self, agent, minimal_config):
        assert agent.config is minimal_config

    def test_security_mode_default(self, agent):
        assert isinstance(agent.security_mode, SecurityMode)

    def test_sessions_manager_created(self, agent):
        assert agent.sessions is not None

    def test_memory_disabled_when_config_says_so(self, agent):
        assert agent.memory is None  # minimal_config disables memory

    def test_scheduler_disabled_when_config_says_so(self, agent):
        assert agent.scheduler is None  # minimal_config disables scheduler

    def test_planner_disabled_when_config_says_so(self, agent):
        assert agent.planner is None

    def test_memory_extractor_disabled_when_config_says_so(self, agent):
        assert agent.memory_extractor is None

    def test_mesh_gateway_none_when_disabled(self, agent):
        assert agent.mesh_gateway is None

    def test_distributed_bus_none_when_disabled(self, agent):
        assert agent.distributed_bus is None

    def test_health_startup_marked_complete(self, agent):
        """Health checker marks startup complete at end of __init__."""
        assert agent.health._startup_complete is True

    def test_custom_security_mode(self, minimal_config, mock_provider):
        """Custom security mode is applied."""
        patches = _agent_patches(mock_provider)
        [p.__enter__() if hasattr(p, "__enter__") else p.start() for p in patches]
        try:
            a = Agent(config=minimal_config, security_mode=SecurityMode.PARANOID)
            assert a.security_mode == SecurityMode.PARANOID
        finally:
            for p in reversed(patches):
                try:
                    if hasattr(p, "__exit__"):
                        p.__exit__(None, None, None)
                except Exception:
                    pass


class TestCreateAgent:
    """create_agent() convenience function."""

    def test_returns_agent(self, minimal_config, mock_provider):
        patches = _agent_patches(mock_provider)
        [p.__enter__() if hasattr(p, "__enter__") else p.start() for p in patches]
        try:
            a = create_agent(minimal_config)
            assert isinstance(a, Agent)
        finally:
            for p in reversed(patches):
                try:
                    if hasattr(p, "__exit__"):
                        p.__exit__(None, None, None)
                except Exception:
                    pass


# ════════════════════════════════════════════════════════════════
# BACKWARD-COMPAT DELEGATES
# ════════════════════════════════════════════════════════════════


class TestProviderManagerDelegates:
    """Properties and methods that delegate to ProviderManager."""

    def test_model_router_property(self, agent):
        agent._provider_manager.model_router = "test_router"
        assert agent.model_router == "test_router"

    def test_model_router_setter(self, agent):
        agent.model_router = "new_router"
        assert agent._provider_manager.model_router == "new_router"

    def test_routing_mode_property(self, agent):
        agent._provider_manager._routing_mode = "manual"
        assert agent._routing_mode == "manual"

    def test_routing_mode_setter(self, agent):
        agent._routing_mode = "auto"
        assert agent._provider_manager._routing_mode == "auto"

    def test_manual_routes_property(self, agent):
        agent._provider_manager._manual_routes = {"chat": "ollama/llama3.2"}
        assert agent._manual_routes == {"chat": "ollama/llama3.2"}

    def test_manual_routes_setter(self, agent):
        agent._manual_routes = {"chat": "gpt-4o"}
        assert agent._provider_manager._manual_routes == {"chat": "gpt-4o"}

    def test_get_routing_config(self, agent):
        agent._provider_manager.get_routing_config = MagicMock(return_value={"mode": "auto"})
        result = agent.get_routing_config()
        assert result == {"mode": "auto"}
        agent._provider_manager.get_routing_config.assert_called_once()

    def test_set_routing_config(self, agent):
        agent._provider_manager.set_routing_config = MagicMock(return_value="ok")
        result = agent.set_routing_config("manual", {"chat": "gpt-4o"})
        assert result == "ok"
        agent._provider_manager.set_routing_config.assert_called_once_with(
            "manual", {"chat": "gpt-4o"}
        )

    def test_switch_provider(self, agent):
        agent._provider_manager.switch_provider = MagicMock(return_value="switched")
        result = agent.switch_provider("openai")
        assert result == "switched"

    def test_get_presets(self, agent):
        agent._provider_manager.get_presets = MagicMock(return_value=[])
        result = agent.get_presets()
        assert result == []

    def test_save_preset(self, agent):
        agent._provider_manager.save_preset = MagicMock(return_value="saved")
        agent.save_preset(1, "My Preset")
        agent._provider_manager.save_preset.assert_called_once_with(1, "My Preset")

    def test_load_preset(self, agent):
        agent._provider_manager.load_preset = MagicMock(return_value="loaded")
        assert agent.load_preset(1) == "loaded"

    def test_set_active_preset(self, agent):
        agent._provider_manager.set_active_preset = MagicMock()
        agent.set_active_preset(2)
        agent._provider_manager.set_active_preset.assert_called_once_with(2)

    def test_rename_preset(self, agent):
        agent._provider_manager.rename_preset = MagicMock()
        agent.rename_preset(1, "New Name")
        agent._provider_manager.rename_preset.assert_called_once_with(1, "New Name")

    def test_delete_preset(self, agent):
        agent._provider_manager.delete_preset = MagicMock()
        agent.delete_preset(3)
        agent._provider_manager.delete_preset.assert_called_once_with(3)


class TestToolExecutorDelegates:
    """Methods that delegate to ToolExecutor."""

    def test_set_tool_event_callback(self, agent):
        cb = MagicMock()
        agent._tool_executor.set_tool_event_callback = MagicMock()
        agent.set_tool_event_callback("user1", cb)
        agent._tool_executor.set_tool_event_callback.assert_called_once_with("user1", cb)

    def test_clear_tool_event_callback(self, agent):
        agent._tool_executor.clear_tool_event_callback = MagicMock()
        agent.clear_tool_event_callback("user1")
        agent._tool_executor.clear_tool_event_callback.assert_called_once_with("user1")

    def test_execute_tool_secure_delegates(self, agent):
        agent._tool_executor.execute_tool_secure = MagicMock(return_value="result")
        result = agent._execute_tool_secure("tool", {})
        assert result == "result"

    def test_execute_tools_delegates(self, agent):
        agent._tool_executor.execute_tools = MagicMock(return_value=([], []))
        result = agent._execute_tools(tool_calls=[])
        assert result == ([], [])

    def test_get_allowed_tools_delegates(self, agent):
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        result = agent._get_allowed_tools(MagicMock())
        assert result == ([], set())

    def test_get_skill_for_tool_delegates(self, agent):
        agent._tool_executor.get_skill_for_tool = MagicMock(return_value="my_skill")
        assert agent._get_skill_for_tool("some_tool") == "my_skill"

    def test_is_tool_error_delegates(self, agent):
        agent._tool_executor._is_tool_error = MagicMock(return_value=True)
        assert agent._is_tool_error("Error: something") is True


class TestMemoryManagerDelegates:
    """Properties and methods that delegate to MemoryManager."""

    def test_episodic_enabled_property(self, agent):
        agent._memory_manager._episodic_enabled = True
        assert agent._episodic_enabled is True

    def test_episodic_enabled_setter(self, agent):
        agent._episodic_enabled = False
        assert agent._memory_manager._episodic_enabled is False

    def test_episodic_dir_property(self, agent):
        test_path = Path("/tmp/test_episodic")
        agent._memory_manager._episodic_dir = test_path
        assert agent._episodic_dir == test_path

    def test_episodic_stores_property(self, agent):
        stores = {"user1": MagicMock()}
        agent._memory_manager._episodic_stores = stores
        assert agent._episodic_stores is stores

    def test_get_episodic_memory_delegates(self, agent):
        agent._memory_manager.get_episodic_memory = MagicMock(return_value="ep_mem")
        assert agent._get_episodic_memory("user1") == "ep_mem"

    def test_save_episodic_memory_async_delegates(self, agent):
        agent._memory_manager.save_episodic_memory_async = MagicMock()
        agent._save_episodic_memory_async("user1")
        agent._memory_manager.save_episodic_memory_async.assert_called_once_with("user1")

    def test_remember_delegates(self, agent):
        agent._memory_manager.remember = MagicMock()
        agent.remember("key", "value", "fact")
        agent._memory_manager.remember.assert_called_once_with("key", "value", "fact")

    def test_recall_delegates(self, agent):
        agent._memory_manager.recall = MagicMock(return_value=["memory1"])
        assert agent.recall("query") == ["memory1"]

    def test_clear_history_delegates(self, agent):
        agent._memory_manager.clear_history = MagicMock()
        agent.clear_history("user1", "telegram")
        agent._memory_manager.clear_history.assert_called_once_with("user1", "telegram")


class TestStatusReporterDelegates:
    """Methods that delegate to StatusReporter."""

    def test_get_status(self, agent):
        agent._status_reporter.get_status = MagicMock(return_value={"name": "test"})
        assert agent.get_status() == {"name": "test"}

    def test_get_health_status(self, agent):
        agent._status_reporter.get_health_status = MagicMock(return_value={"healthy": True})
        assert agent.get_health_status() == {"healthy": True}

    def test_get_health_status_with_resources(self, agent):
        agent._status_reporter.get_health_status = MagicMock()
        agent.get_health_status(include_resources=False)
        agent._status_reporter.get_health_status.assert_called_once_with(False)

    def test_get_metrics_summary(self, agent):
        agent._status_reporter.get_metrics_summary = MagicMock(return_value={"tools": {}})
        result = agent.get_metrics_summary()
        assert result == {"tools": {}}

    def test_get_metrics_summary_params(self, agent):
        agent._status_reporter.get_metrics_summary = MagicMock()
        agent.get_metrics_summary(tool_name="search", include_llm=False)
        agent._status_reporter.get_metrics_summary.assert_called_once_with(
            "search", False, True, True
        )

    def test_get_bus_status(self, agent):
        agent._status_reporter.get_bus_status = MagicMock(return_value={})
        assert agent.get_bus_status() == {}

    def test_get_mesh_status(self, agent):
        agent._status_reporter.get_mesh_status = MagicMock(return_value={})
        assert agent.get_mesh_status() == {}

    def test_get_retriever_status(self, agent):
        agent._status_reporter.get_retriever_status = MagicMock(return_value={})
        assert agent.get_retriever_status() == {}

    def test_mark_ready(self, agent):
        agent._status_reporter.mark_ready = MagicMock()
        agent.mark_ready()
        agent._status_reporter.mark_ready.assert_called_once()

    def test_liveness(self, agent):
        agent._status_reporter.liveness = MagicMock(return_value={"status": "ok"})
        assert agent.liveness() == {"status": "ok"}

    def test_readiness(self, agent):
        agent._status_reporter.readiness = MagicMock(return_value={"ready": True})
        assert agent.readiness() == {"ready": True}

    def test_startup_probe(self, agent):
        agent._status_reporter.startup_probe = MagicMock(return_value={"started": True})
        assert agent.startup_probe() == {"started": True}


class TestPromptBuilderDelegates:
    """Methods that delegate to PromptBuilder."""

    def test_get_system_prompt(self, agent):
        agent._prompt_builder.build_system_prompt = MagicMock(return_value="prompt text")
        result = agent.get_system_prompt(
            session=MagicMock(),
            current_message="hello",
            channel="cli",
        )
        assert result == "prompt text"
        agent._prompt_builder.build_system_prompt.assert_called_once()


# ════════════════════════════════════════════════════════════════
# CHAT ENTRY POINTS
# ════════════════════════════════════════════════════════════════


class TestChat:
    """Agent.chat() synchronous entry point."""

    def test_returns_string(self, agent):
        """chat() returns a plain string response."""
        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            session.session_id = "test-session"
            mock_setup.return_value = (session, "Hello", "user1", None, None)

            # _run_agent_loop is a generator; StopIteration.value is the return
            def fake_loop(**kwargs):
                return iter([])  # empty iterator; value set via StopIteration

            gen = MagicMock()
            gen.__next__ = MagicMock(side_effect=StopIteration("Test response"))
            mock_loop.return_value = gen

            result = agent.chat("Hello")
            assert result == "Test response"

    def test_guardrail_block_returns_message(self, agent):
        """_GuardrailBlock during setup returns user-facing message."""
        with patch.object(agent, "_setup_request") as mock_setup:
            mock_setup.side_effect = _GuardrailBlock("Rate limited")
            result = agent.chat("Hello")
            assert result == "Rate limited"

    def test_passes_user_id(self, agent):
        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            mock_setup.return_value = (session, "Hi", "user42", None, None)
            gen = MagicMock()
            gen.__next__ = MagicMock(side_effect=StopIteration("ok"))
            mock_loop.return_value = gen

            agent.chat("Hi", user_id="user42")
            mock_setup.assert_called_once()
            args = mock_setup.call_args
            assert args[0][1] == "user42"  # user_id arg

    def test_passes_channel(self, agent):
        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            mock_setup.return_value = (session, "Hi", "default", None, None)
            gen = MagicMock()
            gen.__next__ = MagicMock(side_effect=StopIteration("ok"))
            mock_loop.return_value = gen

            agent.chat("Hi", channel="telegram")
            args = mock_setup.call_args
            assert args[0][2] == "telegram"  # channel arg

    def test_loop_called_with_stream_false(self, agent):
        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            mock_setup.return_value = (session, "Hi", "default", None, None)
            gen = MagicMock()
            gen.__next__ = MagicMock(side_effect=StopIteration("ok"))
            mock_loop.return_value = gen

            agent.chat("Hi")
            _, kwargs = mock_loop.call_args
            assert kwargs["stream"] is False
            assert kwargs["emit_tool_events"] is False


class TestChatWithResults:
    """Agent.chat_with_results() returns structured AgentResponse."""

    def test_returns_agent_response(self, agent):
        from familiar.core.agent_response import AgentResponse

        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            mock_setup.return_value = (session, "Hi", "default", None, None)
            gen = MagicMock()
            gen.__next__ = MagicMock(side_effect=StopIteration("response text"))
            mock_loop.return_value = gen

            result = agent.chat_with_results("Hi")
            assert isinstance(result, AgentResponse)
            assert result.text == "response text"

    def test_guardrail_block_returns_agent_response(self, agent):
        from familiar.core.agent_response import AgentResponse

        with patch.object(agent, "_setup_request") as mock_setup:
            mock_setup.side_effect = _GuardrailBlock("Blocked")
            result = agent.chat_with_results("Hi")
            assert isinstance(result, AgentResponse)
            assert result.text == "Blocked"

    def test_exception_returns_error_response(self, agent):
        from familiar.core.agent_response import AgentResponse

        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            mock_setup.return_value = (session, "Hi", "default", None, None)
            gen = MagicMock()
            gen.__next__ = MagicMock(side_effect=RuntimeError("boom"))
            mock_loop.return_value = gen

            result = agent.chat_with_results("Hi")
            assert isinstance(result, AgentResponse)
            assert "error" in result.text.lower()

    def test_passes_job_class_key(self, agent):
        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            mock_setup.return_value = (session, "Hi", "default", None, None)
            gen = MagicMock()
            gen.__next__ = MagicMock(side_effect=StopIteration("ok"))
            mock_loop.return_value = gen

            agent.chat_with_results("Hi", job_class_key="nonprofit_director")
            _, kwargs = mock_loop.call_args
            assert kwargs["job_class_key"] == "nonprofit_director"


class TestChatStream:
    """Agent.chat_stream() streaming entry point."""

    def test_fallback_when_no_streaming(self, agent, mock_provider):
        """Falls back to chat() when provider doesn't support streaming."""
        mock_provider.supports_streaming = False
        with (
            patch.object(agent, "chat", return_value="non-stream response"),
        ):
            events = list(agent.chat_stream("Hello"))
            assert any(e.type == StreamEventType.TEXT for e in events)
            assert any(e.type == StreamEventType.DONE for e in events)

    def test_guardrail_block_yields_text_and_done(self, agent):
        with patch.object(agent, "_setup_request") as mock_setup:
            mock_setup.side_effect = _GuardrailBlock("Rate limited")
            events = list(agent.chat_stream("Hello"))
            texts = [e for e in events if e.type == StreamEventType.TEXT]
            dones = [e for e in events if e.type == StreamEventType.DONE]
            assert len(texts) == 1
            assert texts[0].content == "Rate limited"
            assert len(dones) == 1

    def test_delegates_to_run_agent_loop_with_stream_true(self, agent):
        with (
            patch.object(agent, "_setup_request") as mock_setup,
            patch.object(agent, "_run_agent_loop") as mock_loop,
        ):
            session = MagicMock()
            mock_setup.return_value = (session, "Hi", "default", None, None)
            mock_loop.return_value = iter(
                [
                    StreamEvent(type=StreamEventType.TEXT, content="hello"),
                    StreamEvent(type=StreamEventType.DONE),
                ]
            )

            events = list(agent.chat_stream("Hi"))
            _, kwargs = mock_loop.call_args
            assert kwargs["stream"] is True
            assert kwargs["emit_tool_events"] is True
            assert len(events) == 2


# ════════════════════════════════════════════════════════════════
# _setup_request PRE-FLIGHT
# ════════════════════════════════════════════════════════════════


class TestSetupRequest:
    """_setup_request handles sanitization, sessions, guardrails."""

    def test_returns_five_tuple(self, agent):
        agent.guardrails.start_request = MagicMock(return_value=MagicMock())
        agent.guardrails.process_with_pii = MagicMock(return_value=("Hello", [], False))
        result = agent._setup_request("Hello", "user1", "cli", None)
        assert len(result) == 5
        session, msg, uid, trace, tracker = result
        assert isinstance(session, SecureSession)

    def test_sanitizes_input(self, agent):
        agent.guardrails.start_request = MagicMock(return_value=MagicMock())
        agent.guardrails.process_with_pii = MagicMock(return_value=("cleaned", [], False))
        _, msg, _, _, _ = agent._setup_request("Hello\x00", "user1", "cli", None)
        # Sanitization strips null bytes
        assert "\x00" not in msg

    def test_skips_sanitization_when_flagged(self, agent):
        """already_sanitized=True skips the sanitize_user_input call."""
        agent.guardrails.start_request = MagicMock(return_value=MagicMock())
        agent.guardrails.process_with_pii = MagicMock(return_value=("Hello", [], False))
        with patch("familiar.core.agent.sanitize_user_input") as mock_san:
            agent._setup_request("Hello", "user1", "cli", None, already_sanitized=True)
            mock_san.assert_not_called()

    def test_rate_limit_raises_guardrail_block(self, agent):
        agent.guardrails.start_request = MagicMock(
            side_effect=GuardrailViolation(ViolationType.RATE_LIMIT, "rate limit")
        )
        with pytest.raises(_GuardrailBlock):
            agent._setup_request("Hello", "user1", "cli", None)

    def test_pii_block_raises_guardrail_block(self, agent):
        agent.guardrails.start_request = MagicMock(return_value=MagicMock())
        agent.guardrails.process_with_pii = MagicMock(
            side_effect=GuardrailViolation(ViolationType.CONTENT_SAFETY, "PII detected")
        )
        with pytest.raises(_GuardrailBlock):
            agent._setup_request("my SSN is 123-45-6789", "user1", "cli", None)

    def test_budget_exceeded_raises(self, agent):
        agent.guardrails.start_request = MagicMock(return_value=MagicMock())
        agent.guardrails.process_with_pii = MagicMock(return_value=("Hello", [], False))
        # Make session budget exhausted
        session = agent.sessions.get_or_create_session("user1", "cli")
        session.spent_today = session.daily_budget + 1.0
        with patch.object(agent.sessions, "get_or_create_session", return_value=session):
            with pytest.raises(BudgetExceededError):
                agent._setup_request("Hello", "user1", "cli", None)

    def test_creates_trace_when_tracer_exists(self, agent):
        agent.guardrails.start_request = MagicMock(return_value=MagicMock())
        agent.guardrails.process_with_pii = MagicMock(return_value=("Hello", [], False))
        mock_span = MagicMock()
        agent.tracer = MagicMock()
        agent.tracer.create_span = MagicMock(return_value=mock_span)
        _, _, _, trace, _ = agent._setup_request("Hello", "user1", "cli", None)
        assert trace is mock_span


# ════════════════════════════════════════════════════════════════
# _finalize_response POST-PROCESSING
# ════════════════════════════════════════════════════════════════


class TestFinalizeResponse:
    """_finalize_response handles history, analytics, memory, guardrails."""

    @pytest.fixture(autouse=True)
    def _mock_finalize_deps(self, agent):
        """Mock real objects that _finalize_response interacts with."""
        agent.history = MagicMock()
        agent.sessions = MagicMock()

    def _make_response(self):
        return LLMResponse(
            text="Response",
            tool_calls=[],
            stop_reason="end_turn",
            usage={"input_tokens": 10, "output_tokens": 8},
        )

    def test_saves_history(self, agent):
        response = self._make_response()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)
        result = agent._finalize_response(
            message="Hi",
            final_text="Hello!",
            user_id="user1",
            channel="cli",
            user_context=None,
            session=MagicMock(session_id="s1", audit_log=[]),
            trace=None,
            response=response,
            start_time=time.time(),
        )
        assert isinstance(result, str)
        # History should have 2 new entries (user + assistant)
        assert agent.history.add_message.call_count == 2

    def test_records_analytics(self, agent):
        response = self._make_response()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)
        with patch.object(agent, "_record_analytics") as mock_analytics:
            agent._finalize_response(
                message="Hi",
                final_text="Hello!",
                user_id="user1",
                channel="cli",
                user_context=None,
                session=MagicMock(session_id="s1", audit_log=[]),
                trace=None,
                response=response,
                start_time=time.time(),
            )
            mock_analytics.assert_called_once()

    def test_ends_trace(self, agent):
        response = self._make_response()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)
        mock_trace = MagicMock()
        agent._finalize_response(
            message="Hi",
            final_text="Hello!",
            user_id="user1",
            channel="cli",
            user_context=None,
            session=MagicMock(session_id="s1", audit_log=[]),
            trace=mock_trace,
            response=response,
            start_time=time.time(),
        )
        mock_trace.end.assert_called_once()

    def test_saves_session(self, agent):
        response = self._make_response()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)
        mock_session = MagicMock(session_id="s1", audit_log=[])
        agent._finalize_response(
            message="Hi",
            final_text="Hello!",
            user_id="user1",
            channel="cli",
            user_context=None,
            session=mock_session,
            trace=None,
            response=response,
            start_time=time.time(),
        )
        agent.sessions.save_session.assert_called_once_with(mock_session)

    def test_triggers_memory_extraction(self, agent):
        response = self._make_response()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)
        agent.memory_extractor = MagicMock()
        agent._finalize_response(
            message="Hi",
            final_text="Hello!",
            user_id="user1",
            channel="cli",
            user_context=None,
            session=MagicMock(session_id="s1", audit_log=[]),
            trace=None,
            response=response,
            start_time=time.time(),
        )
        agent.memory_extractor.extract_async.assert_called_once_with("Hi", "Hello!", "cli")

    def test_ends_guardrail_tracking(self, agent):
        response = self._make_response()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)
        tracker = MagicMock()
        agent._finalize_response(
            message="Hi",
            final_text="Hello!",
            user_id="user1",
            channel="cli",
            user_context=None,
            session=MagicMock(session_id="s1", audit_log=[]),
            trace=None,
            response=response,
            start_time=time.time(),
            guardrail_tracker=tracker,
        )
        agent.guardrails.end_request.assert_called_once_with("user1", tracker)

    def test_stores_episodic_memory(self, agent):
        response = self._make_response()
        agent.guardrails.end_request = MagicMock()
        mock_episodic = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=mock_episodic)
        agent._save_episodic_memory_async = MagicMock()
        agent._finalize_response(
            message="Hi",
            final_text="Hello!",
            user_id="user1",
            channel="cli",
            user_context=None,
            session=MagicMock(session_id="s1", audit_log=[]),
            trace=None,
            response=response,
            start_time=time.time(),
        )
        mock_episodic.store_conversation.assert_called_once()
        agent._save_episodic_memory_async.assert_called_once_with("user1")


# ════════════════════════════════════════════════════════════════
# _run_agent_loop PHASE BOUNDARIES
# ════════════════════════════════════════════════════════════════


class TestRunAgentLoop:
    """_run_agent_loop orchestrates history, context, provider, tools."""

    def _make_session(self):
        s = MagicMock()
        s.session_id = "test-session"
        s.remaining_budget = 100.0
        s.charge.return_value = True
        s.audit_log = []
        return s

    def test_no_tool_calls_returns_text(self, agent, mock_provider):
        """When LLM returns text with no tool calls, loop returns it."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)

        gen = agent._run_agent_loop(
            message="Hello",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "Test response" in done.value

    def test_tool_calls_loop_continues(self, agent, mock_provider):
        """When LLM returns tool calls, loop executes them and continues."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)

        # First call returns tool call, second returns text
        tool_call_response = LLMResponse(
            text=None,
            tool_calls=[{"id": "tc1", "name": "search", "input": {"query": "test"}}],
            stop_reason="tool_use",
            usage={"input_tokens": 10, "output_tokens": 5},
        )
        text_response = LLMResponse(
            text="Final answer",
            tool_calls=[],
            stop_reason="end_turn",
            usage={"input_tokens": 15, "output_tokens": 10},
        )
        mock_provider.chat.side_effect = [tool_call_response, text_response]

        agent._tool_executor.execute_tools = MagicMock(
            return_value=(
                [{"tool_use_id": "tc1", "tool_name": "search", "content": "result"}],
                [],
            )
        )

        gen = agent._run_agent_loop(
            message="Search for X",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "Final answer" in done.value
        assert agent._tool_executor.execute_tools.call_count == 1

    def test_max_iterations_returns_timeout(self, agent, mock_provider):
        """Hitting max iterations returns timeout message."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()
        agent.guardrails.end_request = MagicMock()

        # Always return tool calls to force iteration
        mock_provider.chat.return_value = LLMResponse(
            text=None,
            tool_calls=[{"id": "tc1", "name": "search", "input": {}}],
            stop_reason="tool_use",
            usage={"input_tokens": 10, "output_tokens": 5},
        )
        agent._tool_executor.execute_tools = MagicMock(
            return_value=(
                [{"tool_use_id": "tc1", "tool_name": "search", "content": "result"}],
                [],
            )
        )

        gen = agent._run_agent_loop(
            message="Loop forever",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "max iterations" in done.value.lower()

    def test_provider_error_with_fallback(self, agent, mock_provider):
        """LLM error triggers fallback provider."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)

        # First provider fails, fallback succeeds
        mock_provider.chat.side_effect = RuntimeError("API down")
        fallback = MagicMock()
        fallback.name = "FallbackProvider"
        fallback.chat.return_value = LLMResponse(
            text="Fallback response",
            tool_calls=[],
            stop_reason="end_turn",
            usage={"input_tokens": 10, "output_tokens": 8},
        )
        agent._provider_manager.provider_fallback = MagicMock(return_value=fallback)

        gen = agent._run_agent_loop(
            message="Hello",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "Fallback response" in done.value
            assert "unavailable" in done.value.lower() or "using" in done.value.lower()

    def test_provider_error_no_fallback(self, agent, mock_provider):
        """LLM error with no fallback returns error message."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()

        mock_provider.chat.side_effect = RuntimeError("API down")
        agent._provider_manager.provider_fallback = MagicMock(return_value=None)

        gen = agent._run_agent_loop(
            message="Hello",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "unavailable" in done.value.lower()

    def test_phi_guard_blocks_cloud(self, agent, mock_provider):
        """PHI detected blocks cloud provider with appropriate message."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            side_effect=RuntimeError("PHI_NO_LOCAL_FALLBACK")
        )
        agent.guardrails.check_iteration = MagicMock()

        gen = agent._run_agent_loop(
            message="Patient SSN is 123-45-6789",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "health information" in done.value.lower()

    def test_guardrail_iteration_violation(self, agent, mock_provider):
        """Guardrail iteration limit returns clean message."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent.guardrails.check_iteration = MagicMock(
            side_effect=GuardrailViolation(ViolationType.ITERATION_LIMIT, "too many")
        )

        gen = agent._run_agent_loop(
            message="Hello",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "loop" in done.value.lower() or "stuck" in done.value.lower()

    def test_budget_exhausted_mid_loop(self, agent, mock_provider):
        """Budget running out mid-loop returns budget message."""
        session = self._make_session()
        session.charge.return_value = False  # Budget exhausted
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()

        gen = agent._run_agent_loop(
            message="Hello",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration as done:
            assert "budget" in done.value.lower()

    def test_stream_mode_yields_events(self, agent, mock_provider):
        """Stream mode yields StreamEvents."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)

        # Setup streaming — provider returns events via generator
        def fake_stream(**kwargs):
            yield StreamEvent(type=StreamEventType.TEXT, content="Hello ")
            yield StreamEvent(type=StreamEventType.TEXT, content="world")
            return LLMResponse(
                text="Hello world",
                tool_calls=[],
                stop_reason="end_turn",
                usage={"input_tokens": 10, "output_tokens": 5},
            )

        mock_provider.chat_stream = MagicMock(side_effect=fake_stream)

        events = list(
            agent._run_agent_loop(
                message="Hello",
                user_id="user1",
                channel="cli",
                include_history=False,
                user_context=None,
                session=session,
                trace=None,
                stream=True,
                emit_tool_events=True,
            )
        )
        text_events = [e for e in events if e.type == StreamEventType.TEXT]
        done_events = [e for e in events if e.type == StreamEventType.DONE]
        assert len(text_events) >= 1
        assert len(done_events) == 1

    def test_history_included_when_flag_true(self, agent, mock_provider):
        """include_history=True adds conversation history to messages."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)

        # Pre-populate history
        agent.history.get_history = MagicMock(
            return_value=[
                {"role": "user", "content": "Previous message"},
                {"role": "assistant", "content": "Previous response"},
            ]
        )

        gen = agent._run_agent_loop(
            message="New message",
            user_id="user1",
            channel="cli",
            include_history=True,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
        )
        try:
            while True:
                next(gen)
        except StopIteration:
            pass

        # Check that history was passed to provider
        call_kwargs = mock_provider.chat.call_args[1]
        messages = call_kwargs["messages"]
        assert len(messages) == 3  # 2 history + 1 new

    def test_collected_tool_results(self, agent, mock_provider):
        """collected_tool_results accumulates ToolResult objects."""
        session = self._make_session()
        agent._tool_executor.get_allowed_tools = MagicMock(return_value=([], set()))
        agent._prompt_builder.build_system_prompt_parts = MagicMock(
            return_value=("static", "dynamic")
        )
        agent._prompt_builder.inject_memory_tree_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_episodic_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_rag_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_markdown_memory_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_mesh_context = MagicMock(return_value="dynamic")
        agent._prompt_builder.inject_planning_context = MagicMock(return_value="dynamic")
        agent._provider_manager.select_provider = MagicMock(
            return_value=(mock_provider, "mock-model")
        )
        agent.capacity_tracker.is_rate_limited = MagicMock(return_value=False)
        agent.guardrails.check_iteration = MagicMock()
        agent.guardrails.track_llm_usage = MagicMock()
        agent.guardrails.end_request = MagicMock()
        agent._get_episodic_memory = MagicMock(return_value=None)

        tool_call_response = LLMResponse(
            text=None,
            tool_calls=[{"id": "tc1", "name": "search", "input": {"q": "test"}}],
            stop_reason="tool_use",
            usage={"input_tokens": 10, "output_tokens": 5},
        )
        text_response = LLMResponse(
            text="Done",
            tool_calls=[],
            stop_reason="end_turn",
            usage={"input_tokens": 15, "output_tokens": 10},
        )
        mock_provider.chat.side_effect = [tool_call_response, text_response]
        agent._tool_executor.execute_tools = MagicMock(
            return_value=(
                [
                    {
                        "tool_use_id": "tc1",
                        "tool_name": "search",
                        "content": "found it",
                        "tool_input": {"q": "test"},
                    }
                ],
                [],
            )
        )

        collected = []
        gen = agent._run_agent_loop(
            message="Search",
            user_id="user1",
            channel="cli",
            include_history=False,
            user_context=None,
            session=session,
            trace=None,
            stream=False,
            emit_tool_events=False,
            collected_tool_results=collected,
        )
        try:
            while True:
                next(gen)
        except StopIteration:
            pass
        assert len(collected) == 1
        assert collected[0].tool_name == "search"


# ════════════════════════════════════════════════════════════════
# _sync_user_to_session ROLE MAPPING
# ════════════════════════════════════════════════════════════════


class TestSyncUserToSession:
    """_sync_user_to_session maps roles to trust levels."""

    def test_no_user_context_no_op(self, agent):
        session = SecureSession("user1", "cli")
        original_trust = session.trust_level
        agent._sync_user_to_session(None, session)
        assert session.trust_level == original_trust

    def test_owner_telegram_id_promotes(self, agent):
        agent.config.channels = MagicMock()
        agent.config.channels.owner_telegram_id = "12345"
        session = SecureSession("12345", "telegram")
        agent._sync_user_to_session(None, session)
        assert session.trust_level == TrustLevel.OWNER

    def test_persisted_role_restores_trust(self, agent):
        session = SecureSession("user1", "cli")
        session.user_role = "staff"
        session.trust_level = TrustLevel.STRANGER
        agent._sync_user_to_session(None, session)
        assert session.trust_level == TrustLevel.TRUSTED

    def test_admin_role_maps_to_owner(self, agent):
        try:
            from familiar.core.users import UserRole
        except ImportError:
            pytest.skip("users module not available")
        user = MagicMock()
        user.id = "admin1"
        user.email = "admin@test.com"
        user.role = UserRole.ADMIN
        ctx = MagicMock()
        ctx.user = user
        session = SecureSession("admin1", "cli")
        with patch("familiar.core.agent.HAS_USER_MANAGEMENT", True):
            agent._sync_user_to_session(ctx, session)
        assert session.trust_level == TrustLevel.OWNER

    def test_staff_role_maps_to_trusted(self, agent):
        try:
            from familiar.core.users import UserRole
        except ImportError:
            pytest.skip("users module not available")
        user = MagicMock()
        user.id = "staff1"
        user.email = "staff@test.com"
        user.role = UserRole.STAFF
        ctx = MagicMock()
        ctx.user = user
        session = SecureSession("staff1", "cli")
        with patch("familiar.core.agent.HAS_USER_MANAGEMENT", True):
            agent._sync_user_to_session(ctx, session)
        assert session.trust_level == TrustLevel.TRUSTED

    def test_readonly_role_maps_to_known(self, agent):
        try:
            from familiar.core.users import UserRole
        except ImportError:
            pytest.skip("users module not available")
        user = MagicMock()
        user.id = "ro1"
        user.email = "ro@test.com"
        user.role = UserRole.READONLY
        ctx = MagicMock()
        ctx.user = user
        session = SecureSession("ro1", "cli")
        with patch("familiar.core.agent.HAS_USER_MANAGEMENT", True):
            agent._sync_user_to_session(ctx, session)
        assert session.trust_level == TrustLevel.KNOWN


# ════════════════════════════════════════════════════════════════
# GUARDRAIL HANDLING
# ════════════════════════════════════════════════════════════════


class TestGuardrailViolationHandling:
    """_handle_guardrail_violation returns clean user messages."""

    def test_rate_limit_message(self, agent):
        gv = GuardrailViolation(ViolationType.RATE_LIMIT, "too fast")
        msg = agent._handle_guardrail_violation(gv, "user1")
        assert "too quickly" in msg.lower()

    def test_cost_limit_message(self, agent):
        gv = GuardrailViolation(ViolationType.COST_LIMIT, "over budget")
        msg = agent._handle_guardrail_violation(gv, "user1")
        assert "usage limit" in msg.lower()

    def test_token_limit_message(self, agent):
        gv = GuardrailViolation(ViolationType.TOKEN_LIMIT, "too many tokens")
        msg = agent._handle_guardrail_violation(gv, "user1")
        assert "long" in msg.lower()

    def test_iteration_limit_message(self, agent):
        gv = GuardrailViolation(ViolationType.ITERATION_LIMIT, "too many loops")
        msg = agent._handle_guardrail_violation(gv, "user1")
        assert "loop" in msg.lower() or "stuck" in msg.lower()

    def test_timeout_message(self, agent):
        gv = GuardrailViolation(ViolationType.TIMEOUT, "timed out")
        msg = agent._handle_guardrail_violation(gv, "user1")
        assert "long" in msg.lower()

    def test_content_safety_message(self, agent):
        gv = GuardrailViolation(ViolationType.CONTENT_SAFETY, "unsafe")
        msg = agent._handle_guardrail_violation(gv, "user1")
        assert "safety" in msg.lower()

    def test_tool_blocked_message(self, agent):
        gv = GuardrailViolation(ViolationType.TOOL_BLOCKED, "blocked")
        msg = agent._handle_guardrail_violation(gv, "user1")
        assert "admin" in msg.lower()


# ════════════════════════════════════════════════════════════════
# AGENT PROXY
# ════════════════════════════════════════════════════════════════


class TestAgentProxy:
    """AgentProxy sandboxes tool handler access to the agent."""

    def test_exposes_sessions(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.sessions is agent.sessions

    def test_exposes_memory(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.memory is agent.memory

    def test_exposes_tools(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.tools is agent.tools

    def test_exposes_metrics(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.metrics is agent.metrics

    def test_exposes_skills(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.skills is agent.skills

    def test_exposes_guardrails(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.guardrails is agent.guardrails

    def test_exposes_config(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.config is agent.config

    def test_exposes_provider(self, agent):
        proxy = AgentProxy(agent)
        assert proxy.provider is agent.provider

    def test_exposes_name(self, agent):
        proxy = AgentProxy(agent)
        name = proxy.name
        # Should not raise; returns agent config name
        assert name is not None or name is None  # may be mock

    def test_blocks_set_user_trust(self, agent):
        proxy = AgentProxy(agent)
        with pytest.raises(AttributeError):
            proxy.set_user_trust("user1", "cli", TrustLevel.OWNER)

    def test_blocks_grant_capability(self, agent):
        proxy = AgentProxy(agent)
        with pytest.raises(AttributeError):
            proxy.grant_capability("user1", "cli", Capability.HTTP_REQUESTS)

    def test_blocks_deny_capability(self, agent):
        proxy = AgentProxy(agent)
        with pytest.raises(AttributeError):
            proxy.deny_capability("user1", "cli", Capability.HTTP_REQUESTS)

    def test_blocks_internal_managers(self, agent):
        proxy = AgentProxy(agent)
        with pytest.raises(AttributeError):
            _ = proxy._provider_manager
        with pytest.raises(AttributeError):
            _ = proxy._tool_executor
        with pytest.raises(AttributeError):
            _ = proxy._memory_manager

    def test_repr(self, agent):
        proxy = AgentProxy(agent)
        r = repr(proxy)
        assert "AgentProxy" in r


# ════════════════════════════════════════════════════════════════
# BACKUP & HEALTH
# ════════════════════════════════════════════════════════════════


class TestBackup:
    """Backup lifecycle methods."""

    def test_create_backup_delegates(self, agent):
        agent.backup.create_backup = MagicMock()
        mock_manifest = MagicMock()
        mock_manifest.id = "20260317_030000"
        mock_manifest.file_count = 5
        mock_manifest.total_size = 1024
        mock_manifest.status.value = "completed"
        agent.backup.create_backup.return_value = mock_manifest
        result = agent.create_backup()
        assert result is mock_manifest

    def test_restore_backup_delegates(self, agent):
        agent.backup.restore = MagicMock(return_value={"files_restored": 3})
        result = agent.restore_backup("20260317_030000")
        assert result["files_restored"] == 3

    def test_list_backups_delegates(self, agent):
        agent.backup.list_backups = MagicMock(return_value=[])
        assert agent.list_backups() == []

    def test_verify_backup_delegates(self, agent):
        agent.backup.verify_backup = MagicMock(return_value={"valid": True})
        assert agent.verify_backup("id")["valid"] is True

    def test_cleanup_old_backups_delegates(self, agent):
        agent.backup.cleanup_old_backups = MagicMock(return_value=2)
        assert agent.cleanup_old_backups() == 2


class TestHealthChecker:
    """Health checker initialization registers all expected checks."""

    def test_health_checker_created(self, agent):
        assert agent.health is not None

    def test_llm_provider_check_registered(self, agent):
        checks = agent.health._dependency_checks
        assert "llm_provider" in checks

    def test_bus_check_registered(self, agent):
        checks = agent.health._dependency_checks
        assert "bus" in checks

    def test_backup_freshness_check_registered(self, agent):
        checks = agent.health._dependency_checks
        assert "backup_freshness" in checks


# ════════════════════════════════════════════════════════════════
# EVAL DATASET
# ════════════════════════════════════════════════════════════════


class TestEvalDataset:
    """eval export tool and handler."""

    def test_export_disabled_returns_message(self, agent):
        agent._eval_dataset_builder = None
        result = agent.export_eval_dataset()
        assert "not enabled" in result.lower()

    def test_export_no_traces(self, agent):
        agent._eval_dataset_builder = MagicMock()
        agent._eval_dataset_builder.build_dataset.return_value = None
        result = agent.export_eval_dataset("test")
        assert "no traces" in result.lower()

    def test_export_success(self, agent):
        agent._eval_dataset_builder = MagicMock()
        agent._eval_dataset_builder.build_dataset.return_value = "/tmp/eval.json"
        result = agent.export_eval_dataset("test")
        assert "/tmp/eval.json" in result

    def test_handle_export_delegates(self, agent):
        agent._eval_dataset_builder = None
        result = agent._handle_export_eval_dataset({"name": "prod"})
        assert "not enabled" in result.lower()


# ════════════════════════════════════════════════════════════════
# MESH LIFECYCLE
# ════════════════════════════════════════════════════════════════


class TestMeshLifecycle:
    """Mesh gateway lifecycle methods."""

    def test_get_mesh_context_no_gateway(self, agent):
        agent.mesh_gateway = None
        assert agent.get_mesh_context("test query") == ""

    def test_get_mesh_context_no_bridge(self, agent):
        agent.mesh_gateway = MagicMock(spec=[])  # no memory_bridge attr
        assert agent.get_mesh_context("test query") == ""

    def test_get_mesh_context_delegates(self, agent):
        agent.mesh_gateway = MagicMock()
        agent.mesh_gateway.memory_bridge.get_mesh_context.return_value = "peer context"
        result = agent.get_mesh_context("test query")
        assert result == "peer context"

    def test_get_mesh_context_handles_error(self, agent):
        agent.mesh_gateway = MagicMock()
        agent.mesh_gateway.memory_bridge.get_mesh_context.side_effect = RuntimeError("fail")
        assert agent.get_mesh_context("test query") == ""


# ════════════════════════════════════════════════════════════════
# RETRIEVER (RAG)
# ════════════════════════════════════════════════════════════════


class TestRetrieveRagContext:
    """RAG context retrieval for system prompt injection."""

    def test_empty_query(self, agent):
        assert agent._retrieve_rag_context("") == ""

    def test_short_query(self, agent):
        assert agent._retrieve_rag_context("ab") == ""

    def test_no_documents_indexed(self, agent):
        agent.retriever = MagicMock()
        agent.retriever.count = 0
        assert agent._retrieve_rag_context("test query") == ""

    def test_returns_wrapped_context(self, agent):
        agent.retriever = MagicMock()
        agent.retriever.count = 5
        agent.retriever.get_context.return_value = "relevant doc chunk"
        result = agent._retrieve_rag_context("test query")
        assert "<retrieved_knowledge>" in result
        assert "relevant doc chunk" in result

    def test_handles_error_gracefully(self, agent):
        agent.retriever = MagicMock()
        agent.retriever.count = 5
        agent.retriever.get_context.side_effect = RuntimeError("embedding fail")
        assert agent._retrieve_rag_context("test query") == ""


# ════════════════════════════════════════════════════════════════
# SCHEDULER & REMINDERS
# ════════════════════════════════════════════════════════════════


class TestSchedulerHelpers:
    """Scheduler lifecycle and reminder helpers."""

    def test_add_reminder_no_scheduler(self, agent):
        agent.scheduler = None
        result = agent.add_reminder("test", "message", datetime.now(timezone.utc))
        assert result == "Scheduler not enabled"

    def test_start_scheduler_no_scheduler(self, agent):
        agent.scheduler = None
        agent.start_scheduler()  # Should not raise

    def test_stop_scheduler_no_scheduler(self, agent):
        agent.scheduler = None
        agent.stop_scheduler()  # Should not raise


# ════════════════════════════════════════════════════════════════
# SESSION HELPERS
# ════════════════════════════════════════════════════════════════


class TestSessionHelpers:
    """Session management convenience methods."""

    def test_get_session(self, agent):
        session = agent.get_session("user1", "cli")
        assert isinstance(session, SecureSession)

    def test_set_user_trust(self, agent):
        agent.set_user_trust("user1", "cli", TrustLevel.TRUSTED)
        session = agent.get_session("user1", "cli")
        assert session.trust_level == TrustLevel.TRUSTED

    def test_grant_capability(self, agent):
        session = agent.get_session("user1", "cli")
        session.set_trust_level(TrustLevel.TRUSTED)
        agent.grant_capability("user1", "cli", Capability.HTTP_REQUESTS)
        session = agent.get_session("user1", "cli")
        assert Capability.HTTP_REQUESTS.value in session._explicit_grants

    def test_deny_capability(self, agent):
        agent.grant_capability("user1", "cli", Capability.HTTP_REQUESTS)
        agent.deny_capability("user1", "cli", Capability.HTTP_REQUESTS)
        session = agent.get_session("user1", "cli")
        assert Capability.HTTP_REQUESTS.value not in session._explicit_grants


# ════════════════════════════════════════════════════════════════
# SWITCH JOB CLASS
# ════════════════════════════════════════════════════════════════


class TestSwitchJobClass:
    """switch_job_class updates creature state and config."""

    def test_clear_job_class(self, agent):
        with (
            patch("familiar.creature.state.load_creature") as mock_load,
            patch("familiar.creature.state.save_creature"),
        ):
            mock_load.return_value = MagicMock()
            result = agent.switch_job_class("")
            assert "cleared" in result.lower()

    def test_invalid_job_class(self, agent):
        with (
            patch("familiar.jobs.get_job_class", return_value=None),
            patch("familiar.jobs.list_job_classes", return_value=[]),
        ):
            result = agent.switch_job_class("nonexistent")
            assert "unknown" in result.lower()

    def test_valid_job_class(self, agent):
        mock_jc = MagicMock()
        mock_jc.name = "Helper"
        mock_jc.description = "General assistant"
        mock_jc.key = "helper"
        with (
            patch("familiar.jobs.get_job_class", return_value=mock_jc),
            patch("familiar.creature.state.load_creature") as mock_load,
            patch("familiar.creature.state.save_creature"),
        ):
            mock_load.return_value = MagicMock()
            result = agent.switch_job_class("helper")
            assert "Helper" in result
            assert agent.config.agent.job_class == "helper"


# ════════════════════════════════════════════════════════════════
# COMPLIANCE
# ════════════════════════════════════════════════════════════════


class TestCompliance:
    """Compliance mode validation."""

    def test_no_compliance_by_default(self, agent):
        assert agent.compliance_mode is None

    def test_hipaa_requires_encryption_key(self, minimal_config, mock_provider):
        """HIPAA mode without encryption key raises RuntimeError."""
        minimal_config.compliance = MagicMock()
        minimal_config.compliance.mode = "hipaa"
        patches = _agent_patches(mock_provider)
        [p.__enter__() if hasattr(p, "__enter__") else p.start() for p in patches]
        try:
            with pytest.raises(RuntimeError, match="encryption"):
                Agent(config=minimal_config)
        finally:
            for p in reversed(patches):
                try:
                    if hasattr(p, "__exit__"):
                        p.__exit__(None, None, None)
                except Exception:
                    pass


# ════════════════════════════════════════════════════════════════
# ERROR COMPAT
# ════════════════════════════════════════════════════════════════


class TestErrorCompat:
    """Backward-compat error helpers."""

    def test_error_prefixes_accessible(self, agent):
        assert hasattr(Agent, "_ERROR_PREFIXES")
        assert isinstance(Agent._ERROR_PREFIXES, (list, tuple, frozenset, set))

    def test_error_substrings_accessible(self, agent):
        assert hasattr(Agent, "_ERROR_SUBSTRINGS")
        assert isinstance(Agent._ERROR_SUBSTRINGS, (list, tuple, frozenset, set))

    def test_is_tool_error_delegates(self, agent):
        agent._tool_executor._is_tool_error = MagicMock(return_value=True)
        assert agent._is_tool_error("Error: test") is True


# ════════════════════════════════════════════════════════════════
# GUARDRAIL BLOCK EXCEPTION
# ════════════════════════════════════════════════════════════════


class TestGuardrailBlockException:
    """_GuardrailBlock is an internal sentinel."""

    def test_is_exception(self):
        assert issubclass(_GuardrailBlock, Exception)

    def test_carries_message(self):
        gb = _GuardrailBlock("Rate limited")
        assert str(gb) == "Rate limited"

    def test_does_not_propagate_through_chat(self, agent):
        with patch.object(agent, "_setup_request") as mock_setup:
            mock_setup.side_effect = _GuardrailBlock("blocked")
            result = agent.chat("test")
            assert result == "blocked"
